import numpy
from typing import Union, Any, Iterable

NoneInt = Union[int, None]
NumpyInt = Union[
    numpy.int32, # Preferred
    int, # Preferred
    numpy.int64,
    numpy.int16,
    numpy.int8,
]

class Stack:
    def __init__(self, max_size: NoneInt=None) -> None: ...
    def push(self, item: Any) -> None: ...
    def clear(self) -> None: ...

    def pop(self) -> Any: ...
    def peek(self) -> Any: ...

    def is_empty(self) -> bool: ...
    def is_full(self) -> bool: ...
    def changed(self) -> bool: ...

    def size(self) -> int: ...

class Queue:
    def __init__(self, max_size: NoneInt=None) -> None: ...
    def enqueue(self, item: Any) -> None: ...
    def clear(self) -> None: ...

    def dequeue(self) -> Any: ...
    def peek(self) -> Any: ...

    def is_empty(self) -> bool: ...
    def is_full(self) -> bool: ...
    def changed(self) -> bool: ...

    def size(self) -> int: ...

class CircularQueue:
    def __init__(self, size: int) -> None: ...
    def clear(self) -> None: ...
    def enqueue(self, item: Any) -> None: ...

    def dequeue(self) -> Any: ...
    def peek(self) -> Any: ...

    def size(self) -> int: ...

    def is_empty(self) -> bool: ...
    def is_full(self) -> bool: ...
    def changed(self) -> bool: ...

class PriorityQueue:
    def __init__(self) -> None: ...
    def enqueue(self, item: Any, priority: NumpyInt) -> None: ...
    def clear(self) -> None: ...

    def dequeue(self) -> Any: ...
    def peek(self) -> Any: ...

    def peek_next_priority(self) -> NumpyInt: ...
    def size(self) -> int: ...

    def is_empty(self) -> bool: ...
    def changed(self) -> bool: ...

class InvertedPriorityQueue:
    def __init__(self) -> None: ...
    def enqueue(self, item: Any, priority: NumpyInt) -> None: ...
    def clear(self) -> None: ...

    def dequeue(self) -> Any: ...
    def peek(self) -> Any: ...

    def peek_next_priority(self) -> NumpyInt: ...
    def size(self) -> int: ...

    def is_empty(self) -> bool: ...
    def changed(self) -> bool: ...

class PriorityList:
    def __init__(self) -> None: ...
    def add(self, item: Any, priority: NumpyInt) -> None: ...
    def remove_item(self, item: Any) -> None: ...
    def clear(self) -> None: ...
    def update_priority(self, item: Any, new_priority: NumpyInt) -> None: ...

    def remove_highest_priority(self) -> Iterable[Any]: ...

    def peek(self) -> Any: ...

    def peek_next_priority(self) -> NumpyInt: ...
    def size(self) -> int: ...

    def is_empty(self) -> bool: ...
    def changed(self) -> bool: ...

class InvertedPriorityList:
    def __init__(self) -> None: ...
    def add(self, item: Any, priority: NumpyInt) -> None: ...
    def update_priority(self, item: Any, new_priority: NumpyInt) -> None: ...
    def remove_item(self, item: Any) -> None: ...
    def clear(self) -> None: ...

    def remove_highest_priority(self) -> Iterable[Any]: ...

    def peek(self) -> Any: ...

    def peek_next_priority(self) -> NumpyInt: ...
    def size(self) -> int: ...

    def is_empty(self) -> bool: ...
    def changed(self) -> bool: ...


from typing import Iterable, Union

import numpy as np
import numpy.typing as npt

from pmma.build.NumberFormats import Color, DisplayCoordinate

Numerical1D = Union[
    npt.NDArray[np.float32], # preferred
    npt.NDArray[np.float16],
    npt.NDArray[np.float64],
    npt.NDArray[np.int32], # preferred
    npt.NDArray[np.int8],
    npt.NDArray[np.int16],
    npt.NDArray[np.int64],
    Iterable[float],  # preferred
    Iterable[int],  # preferred
]

Integer1D = Union[
    npt.NDArray[np.int32], # preferred
    npt.NDArray[np.int8],
    npt.NDArray[np.int16],
    npt.NDArray[np.int64],
    Iterable[int],
]

Integer2D = Union[
    npt.NDArray[np.int32], # preferred
    npt.NDArray[np.int8],
    npt.NDArray[np.int16],
    npt.NDArray[np.int64],
    Iterable[Iterable[int]],
]

Float1D = Union[
    npt.NDArray[np.float32], # preferred
    npt.NDArray[np.float16],
    npt.NDArray[np.float64],
    Iterable[float]]

Numerical = Union[
    float, int
]

NoneInt = Union[int, None]

class RadialPolygon:
    def __init__(self) -> None: ...

    shape_center: DisplayCoordinate
    shape_color: Color

    def render(self) -> None: ...

    def set_radius(self, radius: int) -> None: ...
    def set_point_count(self, point_count: int) -> None: ...
    def set_width(self, width: int) -> None: ...
    def set_rotation(self, rotation: Numerical) -> None: ...

    def get_radius(self) -> int: ...
    def get_point_count(self) -> int: ...
    def get_width(self) -> int: ...
    def get_rotation(self) -> Numerical: ...

class Rectangle:
    def __init__(self) -> None: ...

    shape_center: DisplayCoordinate
    shape_color: Color

    def render(self) -> None: ...

    def set_size(self, size: Integer1D) -> None: ...
    def set_corner_radius(self, radius: int) -> None: ...
    def set_width(self, width: int) -> None: ...
    def set_rotation(self, rotation: Numerical) -> None: ...

    def get_size(self) -> Integer1D: ...
    def get_corner_radius(self) -> int: ...
    def get_width(self) -> int: ...
    def get_rotation(self) -> Numerical: ...

class Pixel:
    def __init__(self) -> None: ...

    shape_center: DisplayCoordinate
    shape_color: Color

    def render(self) -> None: ...

class Line:
    def __init__(self) -> None: ...

    shape_color: Color
    shape_start: DisplayCoordinate
    shape_end: DisplayCoordinate

    def render(self) -> None: ...

    def set_width(self, width: int) -> None: ...
    def set_rotation(self, rotation: Numerical) -> None: ...

    def get_width(self) -> int: ...
    def get_rotation(self) -> Numerical: ...

class PolygonShape:
    def __init__(self) -> None: ...

    shape_color: Color

    def render(self) -> None: ...

    def set_points(self, points: Integer2D) -> None: ...
    def set_width(self, width: int) -> None: ...
    def set_rotation(self, rotation: Numerical) -> None: ...
    def set_closed(self, closed: bool) -> None: ...

    def get_points(self) -> Integer2D: ...
    def get_point_count(self) -> int: ...
    def get_width(self) -> int: ...
    def get_rotation(self) -> Numerical: ...
    def get_closed(self) -> bool: ...

class Arc:
    def __init__(self) -> None: ...

    shape_center: DisplayCoordinate
    shape_color: Color

    def render(self) -> None: ...

    def set_radius(self, radius: Numerical) -> None: ...
    def set_point_count(self, point_count: int) -> None: ...
    def set_width(self, width: Numerical) -> None: ...
    def set_rotation(self, rotation: Numerical) -> None: ...
    def set_start_angle(self, start_angle: Numerical) -> None: ...
    def set_end_angle(self, end_angle: Numerical) -> None: ...

    def get_radius(self) -> Numerical: ...
    def get_point_count(self) -> int: ...
    def get_width(self) -> Numerical: ...
    def get_rotation(self) -> Numerical: ...
    def get_start_angle(self) -> Numerical: ...
    def get_end_angle(self) -> Numerical: ...

class Ellipse:
    def __init__(self) -> None: ...

    shape_center: DisplayCoordinate
    shape_color: Color

    def render(self) -> None: ...

    def set_size(self, size: Integer1D) -> None: ...
    def set_point_count(self, point_count: int) -> None: ...
    def set_width(self, width: int) -> None: ...
    def set_rotation(self, rotation: Numerical) -> None: ...

    def get_size(self) -> Integer1D: ...
    def get_point_count(self) -> int: ...
    def get_width(self) -> int: ...
    def get_rotation(self) -> Numerical: ...

# Complex shapes

class Circle:
    def __init__(self) -> None: ...

    shape_center: DisplayCoordinate
    shape_color: Color

    def render(self) -> None: ...

    def set_radius(self, radius: int) -> None: ...
    def set_width(self, width: int) -> None: ...

class EquilateralTriangle:
    def __init__(self) -> None: ...

    shape_center: DisplayCoordinate
    shape_color: Color

    def render(self) -> None: ...

    def set_radius(self, radius: int) -> None: ...
    def set_width(self, width: int) -> None: ...
    def set_rotation(self, rotation: Numerical) -> None: ...

class RegularSquare:
    def __init__(self) -> None: ...

    shape_color: Color

    def render(self) -> None: ...

    def set_size(self, size: Numerical) -> None: ...
    def set_corner_radius(self, radius: int) -> None: ...
    def set_width(self, width: int) -> None: ...
    def set_rotation(self, rotation: Numerical) -> None: ...

class RegularPentagon:
    def __init__(self) -> None: ...

    shape_center: DisplayCoordinate
    shape_color: Color

    def render(self) -> None: ...

    def set_radius(self, radius: int) -> None: ...
    def set_width(self, width: int) -> None: ...
    def set_rotation(self, rotation: Numerical) -> None: ...

class RegularHexagon:
    def __init__(self) -> None: ...

    shape_center: DisplayCoordinate
    shape_color: Color

    def render(self) -> None: ...

    def set_radius(self, radius: int) -> None: ...
    def set_width(self, width: int) -> None: ...
    def set_rotation(self, rotation: Numerical) -> None: ...

class RegularHeptagon:
    def __init__(self) -> None: ...

    shape_center: DisplayCoordinate
    shape_color: Color

    def render(self) -> None: ...

    def set_radius(self, radius: int) -> None: ...
    def set_width(self, width: int) -> None: ...
    def set_rotation(self, rotation: Numerical) -> None: ...

class RegularOctagon:
    def __init__(self) -> None: ...

    shape_center: DisplayCoordinate
    shape_color: Color

    def render(self) -> None: ...

    def set_radius(self, radius: int) -> None: ...
    def set_width(self, width: int) -> None: ...
    def set_rotation(self, rotation: Numerical) -> None: ...

class RegularNonagon:
    def __init__(self) -> None: ...

    shape_center: DisplayCoordinate
    shape_color: Color

    def render(self) -> None: ...

    def set_radius(self, radius: int) -> None: ...
    def set_width(self, width: int) -> None: ...
    def set_rotation(self, rotation: Numerical) -> None: ...

class RegularDecagon:
    def __init__(self) -> None: ...

    shape_center: DisplayCoordinate
    shape_color: Color

    def render(self) -> None: ...

    def set_radius(self, radius: int) -> None: ...
    def set_width(self, width: int) -> None: ...
    def set_rotation(self, rotation: Numerical) -> None: ...


from typing import Iterable, Union

Numerical = Union[float, int]

class TextInput:
    def __init__(self) -> None: ...

    def get_text(self) -> str: ...

    def get_enabled(self) -> bool: ...
    def set_enabled(self, value: bool) -> None: ...

    def clear_text(self) -> None: ...

class FileDropped:
    def get_file_paths(self) -> Iterable[str]: ...
    def get_file_paths_toggle(self) -> Iterable[str]: ...

    def get_file_path_count(self) -> int: ...

    def clear_file_paths(self) -> None: ...

    def get_enabled(self) -> bool: ...
    def set_enabled(self, enabled: bool) -> None: ...


import threading
import queue
import time

import soundfile
import waiting
import numpy
import pedalboard

try:
    import sounddevice
    _sounddevice_available = True
    _sounddevice_error = ""
except Exception as _error:
    _sounddevice_available = False

    _sounddevice_error = str(_error)

from pmma.build.NumberFormats import Proportion, LinkedProportion

from pmma.core.py_src.Constants import Constants

class Audio:
    def __init__(self):
        if (_sounddevice_available == False):
            print(f"Sounddevice is not available because ({_sounddevice_error}), audio playback is not \
possible. To fix do: 'sudo apt install libportaudio2' or consult our \
troubleshooting section here: \
https://github.com/PycraftDeveloper/PMMA/blob/main/repo/Troubleshooting/linux.md#oserror-portaudio-library-not-found")

        self._file = None
        self._sample_rate = None
        self._audio_loaded = False
        self._effects_list = []
        self._paused = False
        self._stop_signal = False
        self._playback_thread = None
        self._channels = 2
        self._queue_max_size = 60
        self._audio_queue = queue.Queue(maxsize=self._queue_max_size)
        self._audio_data = None
        self._from_moviepy = False
        self._moviepy_audio_itr = None

        self.volume = Proportion()  # Default volume is 100%
        self.volume.set_proportion_decimal(1.0)
        self.pan = Proportion()  # Pan: -1 (left) to 1 (right), 0 is center
        self.pan.set_proportion_decimal(0.0)

        self._playing = False
        self._first_run = True

        self._looping = False
        self._audio_duration = None
        self._audio_playing_start_time = None

        self._custom_audio_callback_pre_effects = None
        self._custom_audio_callback_post_effects = None

        self._custom_audio_callback_pre_effects_results = None
        self._custom_audio_callback_post_effects_results = None

    def __del__(self):
        """
        🟩 **R** -
        """
        self._stop_signal = True

        if self._playback_thread is not None:
            self._playback_thread.join()

    def set_custom_audio_callback_function_pre_effects(self, function=None):
        """
        🟩 **R** -
        """
        self._custom_audio_callback_pre_effects = function

    def set_custom_audio_callback_function_post_effects(self, function=None):
        """
        🟩 **R** -
        """
        self._custom_audio_callback_post_effects = function

    def get_custom_audio_callback_function_pre_effects_results(self):
        """
        🟩 **R** -
        """
        return self._custom_audio_callback_pre_effects_results

    def get_custom_audio_callback_function_post_effects_results(self):
        """
        🟩 **R** -
        """
        return self._custom_audio_callback_post_effects_results

    def load_from_file(self, file_path):
        """
        🟩 **R** -
        """
        self._file = soundfile.SoundFile(file_path)
        self._sample_rate = self._file.samplerate
        self._channels = self._file.channels
        self._audio_duration = self._file.frames / self._sample_rate
        self._audio_loaded = True

    def get_sample_rate(self):
        """
        🟩 **R** -
        """
        return self._sample_rate

    def get_number_of_channels(self):
        """
        🟩 **R** -
        """
        return self._channels

    def set_looping(self, loop):
        """
        🟩 **R** -
        """
        self._looping = loop

    def get_looping(self):
        """
        🟩 **R** -
        """
        return self._looping

    def set_duration(self, duration):
        """
        🟩 **R** -
        """
        self._audio_duration = duration

    def set_silence_at_end_of_track(self, duration):
        """
        🟩 **R** -
        """
        self._audio_duration += duration

    def load_from_moviepy(self, audio):
        """
        🟩 **R** -
        """
        self._audio_data = audio
        self._sample_rate = audio.fps
        self._channels = audio.nchannels
        self._audio_loaded = True
        self._from_moviepy = True
        self._moviepy_audio_itr = self._audio_generator(2048)

        self._looping = True

        # Pre-fill the queue with initial chunks
        for _ in range(self._queue_max_size):
            try:
                self._audio_queue.put_nowait(next(self._moviepy_audio_itr))
            except StopIteration:
                break

    def add_effect(self, effect):
        """
        🟩 **R** -
        """
        self._effects_list.append(effect)

    def remove_effect(self, effect):
        """
        🟩 **R** -
        """
        self._effects_list.remove(effect)

    def play(self, blocking=True, delay=0):
        """
        🟩 **R** -
        """
        if self._audio_loaded:
            if self._playing:
                print("You have attempted to play audio \
that's already playing. We will therefore ignore your request to prevent unexpected audio behavior.")

                return False
            if delay != 0:
                time.sleep(delay)
            self._effects = pedalboard.Pedalboard(self._effects_list)
            self._paused = False
            self._stop_signal = False
            self._first_run = True

            self._playing = True

            if blocking:
                # Start playback in the current thread (blocking)
                self._start_playback()
            else:
                # Start playback in a separate thread (non-blocking)
                self._playback_thread = threading.Thread(target=self._start_playback)
                self._playback_thread.daemon = True
                self._playback_thread.name = "Audio:Playing_Audio_Thread"
                self._playback_thread.start()
            return True

    def _wait_for_chunk_to_play(self):
        """
        🟩 **R** -
        """
        return self._stop_signal

    def _start_playback(self):
        """
        🟩 **R** -
        """
        if (_sounddevice_available == False):
            print("Sounddevice is not available, ignoring")
            self._audio_playing_start_time = None
            self._file.seek(0)
            return

        self._file.seek(0)
        # Start the audio stream
        with sounddevice.OutputStream(callback=self._audio_callback, samplerate=self._sample_rate, channels=self._channels, blocksize=2048):
            # Loop while playback is ongoing and not stopped
            waiting.wait(self._wait_for_chunk_to_play)

        self._audio_playing_start_time = None
        self._file.seek(0)

    def _audio_generator(self, chunk_size):
        """
        🟩 **R** -
        """
        buffer = numpy.empty((0, self._channels), dtype='float32')  # Buffer to store leftover samples

        while self._stop_signal is False:
            for chunk in self._audio_data.iter_chunks(fps=self._sample_rate, chunksize=chunk_size):
                # Add the new chunk to the buffer
                buffer = numpy.vstack([buffer, chunk])

                # Keep yielding exact-sized chunks from the buffer
                while len(buffer) >= chunk_size:
                    # Yield a chunk of the requested size
                    yield buffer[:chunk_size]
                    # Remove the yielded chunk from the buffer
                    buffer = buffer[chunk_size:]

            # If stop signal is raised or no more chunks, yield remaining data in buffer
            if len(buffer) > 0:
                yield buffer
                break

    def _audio_callback(self, outdata, frames, _, status):
        """
        🟩 **R** -
        """
        if self._audio_playing_start_time is None:
            self._audio_playing_start_time = time.perf_counter()

        if status:
            print(status)

        if self._paused or self._stop_signal or status:
            outdata[:] = numpy.zeros(outdata.shape)
            if self._custom_audio_callback_pre_effects is not None:
                self._custom_audio_callback_pre_effects_results = self._custom_audio_callback_pre_effects(outdata, frames, time, status)
                self._custom_audio_callback_post_effects_results = self._custom_audio_callback_post_effects(outdata, frames, time, status)
            return

        if self._from_moviepy:
            try:
                chunk = self._audio_queue.get_nowait()
                self._audio_queue.put_nowait(next(self._moviepy_audio_itr))
            except StopIteration:
                if self._looping:
                    self._moviepy_audio_itr = self._audio_generator(2048)
                    self._audio_queue.put_nowait(next(self._moviepy_audio_itr))
            except queue.Empty:
                if self._looping:
                    outdata.fill(0)
                else:
                    self._stop_signal = True
                    outdata[:] = numpy.zeros(outdata.shape)
                    if self._custom_audio_callback_pre_effects is not None:
                        self._custom_audio_callback_pre_effects_results = self._custom_audio_callback_pre_effects(outdata, frames, time, status)
                        self._custom_audio_callback_post_effects_results = self._custom_audio_callback_post_effects(outdata, frames, time, status)
                    return

        else:
            chunk = self._file.read(frames, dtype='float32')

            if time.perf_counter() - self._audio_playing_start_time > self._audio_duration:
                if self._looping:
                    self._file.seek(0)
                    self._audio_playing_start_time = None
                else:
                    self._stop_signal = True

        if len(chunk) < frames:
            padding_shape = (frames - len(chunk), chunk.shape[1])
            chunk = numpy.pad(chunk, ((0, padding_shape[0]), (0, 0)), mode='constant')

        if self._custom_audio_callback_pre_effects is not None:
            self._custom_audio_callback_pre_effects_results = self._custom_audio_callback_pre_effects(outdata, frames, time, status)

        # Apply volume and panning
        chunk = self._apply_volume_and_pan(chunk)

        # Apply effects
        processed_audio = self._effects(chunk, self._sample_rate, reset=self._first_run)

        # Output the processed audio
        outdata[:] = processed_audio

        if self._custom_audio_callback_post_effects is not None:
            self._custom_audio_callback_post_effects_results = self._custom_audio_callback_post_effects(outdata, frames, time, status)

        self._first_run = False

    def _apply_volume_and_pan(self, chunk):
        """
        🟩 **R** - Apply volume and panning to the chunk of audio
        """
        chunk = chunk * self.volume.get_proportion_decimal()

        if self._channels == 2:  # Stereo audio
            left = 1 - max(self.pan.get_proportion_decimal(), 0)  # Reduce left channel when panning right
            right = 1 - max(-self.pan.get_proportion_decimal(), 0)  # Reduce right channel when panning left
            chunk[:, 0] *= left
            chunk[:, 1] *= right
        return chunk

    def pause(self):
        """
        🟩 **R** -
        """
        self._paused = True

    def resume(self):
        """
        🟩 **R** -
        """
        if self._paused:
            self._paused = False

    def stop(self):
        """
        🟩 **R** -
        """
        self._stop_signal = True
        if self._playback_thread is not None: # wait until playback actually stops
            self._playback_thread.join()
        self._playing = False

    def get_paused(self):
        """
        🟩 **R** -
        """
        return self._paused

    def get_playing(self):
        """
        🟩 **R** -
        """
        return self._playing

class Chorus(pedalboard.Chorus):
    """
    🟩 **R** -
    """
    def __init__(
            self,
            rate=1,
            depth=25,
            center_delay=0.007,
            feedback=0,
            mix=50,
            format=Constants.PERCENTAGE,
            depth_format=None,
            feedback_format=None,
            mix_format=None):
        """
        🟩 **R** -
        """

        if depth_format is None:
            depth_format = format
        if feedback_format is None:
            feedback_format = format
        if mix_format is None:
            mix_format = format

        self.proportion_adjusted_depth = LinkedProportion(self, "depth")
        if depth_format == Constants.PERCENTAGE:
            self.proportion_adjusted_depth.set_proportion_percentage(depth)
        else:
            self.proportion_adjusted_depth.set_proportion_decimal(depth)

        self.proportion_adjusted_feedback = LinkedProportion(self, "feedback")

        if feedback_format == Constants.PERCENTAGE:
            self.proportion_adjusted_feedback.set_proportion_percentage(feedback)
        else:
            self.proportion_adjusted_feedback.set_proportion_decimal(feedback)

        self.proportion_adjusted_mix = LinkedProportion(self, "mix")
        if mix_format == Constants.PERCENTAGE:
            self.proportion_adjusted_mix.set_proportion_percentage(mix)
        else:
            self.proportion_adjusted_mix.set_proportion_decimal(mix)

        super().__init__(
            rate_hz=rate,
            depth=self.proportion_adjusted_depth.get_proportion_decimal(),
            center_delay_ms=center_delay * 1000, # s to ms
            feedback=self.proportion_adjusted_feedback.get_proportion_decimal(),
            mix=self.proportion_adjusted_mix.get_proportion_decimal())

    def set_rate(self, rate):
        """
        🟩 **R** -
        """
        self.rate_hz = rate

    def get_rate(self):
        """
        🟩 **R** -
        """
        return self.rate_hz

    def set_center_delay(self, center_delay):
        """
        🟩 **R** -
        """
        self.center_delay_ms = center_delay * 1000

    def get_center_delay(self):
        """
        🟩 **R** -
        """
        return self.center_delay_ms / 1000

class Compressor(pedalboard.Compressor):
    """
    🟩 **R** -
    """
    def __init__(
            self,
            threshold=0,
            ratio=4,
            attack=0.001,
            release=0.1):
        """
        🟩 **R** -
        """

        super().__init__(
            threshold_db=threshold,
            ratio=ratio,
            attack_ms=attack * 1000,
            release_ms=release * 1000)

    def set_threshold(self, threshold):
        """
        🟩 **R** -
        """
        self.threshold_db = threshold

    def get_threshold(self):
        """
        🟩 **R** -
        """
        return self.threshold_db

    def set_ratio(self, ratio):
        """
        🟩 **R** -
        """
        self.ratio = ratio

    def get_ratio(self):
        """
        🟩 **R** -
        """
        return self.ratio

    def set_attack(self, attack):
        """
        🟩 **R** -
        """
        self.attack_ms = attack * 1000

    def get_attack(self):
        """
        🟩 **R** -
        """
        return self.attack_ms / 1000

    def set_release(self, release):
        """
        🟩 **R** -
        """
        self.release_ms = release * 1000

    def get_release(self):
        """
        🟩 **R** -
        """
        return self.release_ms / 1000

class Convolution(pedalboard.Convolution):
    """
    🟩 **R** -
    """
    def __init__(
            self,
            impulse_response_filename,
            mix=100,
            sample_rate=None,
            format=Constants.PERCENTAGE,
            mix_format=None):
        """
        🟩 **R** -
        """

        if mix_format is None:
            mix_format = format

        self.proportion_adjusted_mix = LinkedProportion(self, "mix")

        if mix_format == Constants.PERCENTAGE:
            self.proportion_adjusted_mix.set_proportion_percentage(mix)
        else:
            self.proportion_adjusted_mix.set_proportion_decimal(mix)

        super().__init__(
            impulse_response_filename=impulse_response_filename,
            mix=self.proportion_adjusted_mix.get_proportion_decimal(),
            sample_rate=sample_rate)

    def set_impulse_response_filename(self, impulse_response):
        """
        🟩 **R** -
        """
        self.impulse_response = impulse_response

    def get_impulse_response_filename(self):
        """
        🟩 **R** -
        """
        return self.impulse_response

    def set_sample_rate(self, sample_rate):
        """
        🟩 **R** -
        """
        self.sample_rate = sample_rate

    def get_sample_rate(self):
        """
        🟩 **R** -
        """
        return self.sample_rate

class Delay(pedalboard.Delay):
    """
    🟩 **R** -
    """
    def __init__(
            self,
            delay=0.5,
            feedback=0,
            mix=50,
            format=Constants.PERCENTAGE,
            feedback_format=None,
            mix_format=None):
        """
        🟩 **R** -
        """

        if feedback_format is None:
            feedback_format = format
        if mix_format is None:
            mix_format = format

        self.proportion_adjusted_feedback = LinkedProportion(self, "feedback")

        if feedback_format == Constants.PERCENTAGE:
            self.proportion_adjusted_feedback.set_proportion_percentage(feedback)
        else:
            self.proportion_adjusted_feedback.set_proportion_decimal(feedback)

        self.proportion_adjusted_mix = LinkedProportion(self, "mix")
        if mix_format == Constants.PERCENTAGE:
            self.proportion_adjusted_mix.set_proportion_percentage(mix)
        else:
            self.proportion_adjusted_mix.set_proportion_decimal(mix)

        super().__init__(
            delay_seconds=delay,
            feedback=self.proportion_adjusted_feedback.get_proportion_decimal(),
            mix=self.proportion_adjusted_mix.get_proportion_decimal())

    def set_delay(self, delay):
        """
        🟩 **R** -
        """
        self.delay_seconds = delay

    def get_delay(self):
        """
        🟩 **R** -
        """
        return self.delay_seconds

class Phaser(pedalboard.Phaser):
    """
    🟩 **R** -
    """
    def __init__(
            self,
            rate=1,
            depth=50,
            center_frequency=1300,
            feedback=0,
            mix=50,
            format=Constants.PERCENTAGE,
            depth_format=None,
            feedback_format=None,
            mix_format=None):
        """
        🟩 **R** -
        """

        if depth_format is None:
            depth_format = format
        if feedback_format is None:
            feedback_format = format
        if mix_format is None:
            mix_format = format

        self.proportion_adjusted_depth = LinkedProportion(self, "depth")
        if depth_format == Constants.PERCENTAGE:
            self.proportion_adjusted_depth.set_proportion_percentage(depth)
        else:
            self.proportion_adjusted_depth.set_proportion_decimal(depth)

        self.proportion_adjusted_feedback = LinkedProportion(self, "feedback")
        if feedback_format == Constants.PERCENTAGE:
            self.proportion_adjusted_feedback.set_proportion_percentage(feedback)
        else:
            self.proportion_adjusted_feedback.set_proportion_decimal(feedback)

        self.proportion_adjusted_mix = LinkedProportion(self, "mix")
        if mix_format == Constants.PERCENTAGE:
            self.proportion_adjusted_mix.set_proportion_percentage(mix)
        else:
            self.proportion_adjusted_mix.set_proportion_decimal(mix)

        super().__init__(
            rate_hz=rate,
            depth=self.proportion_adjusted_depth.get_proportion_decimal(),
            center_frequency_hz=center_frequency,
            feedback=self.proportion_adjusted_feedback.get_proportion_decimal(),
            mix=self.proportion_adjusted_mix.get_proportion_decimal())

    def set_rate(self, rate):
        """
        🟩 **R** -
        """
        self.rate_hz = rate

    def get_rate(self):
        """
        🟩 **R** -
        """
        return self.rate_hz

    def set_center_frequency(self, center_frequency):
        """
        🟩 **R** -
        """
        self.center_frequency_hz = center_frequency

    def get_center_frequency(self):
        """
        🟩 **R** -
        """
        return self.center_frequency_hz

class Reverb(pedalboard.Reverb):
    """
    🟩 **R** -
    """
    def __init__(
            self,
            room_size=100,
            damping=50,
            wet_level=33,
            dry_level=40,
            width=100,
            freeze_mode=False,
            format=Constants.PERCENTAGE,
            room_size_format=None,
            damping_format=None,
            wet_level_format=None,
            dry_level_format=None,
            width_format=None):
        """
        🟩 **R** -
        """

        if room_size_format is None:
            room_size_format = format
        if damping_format is None:
            damping_format = format
        if wet_level_format is None:
            wet_level_format = format
        if dry_level_format is None:
            dry_level_format = format
        if width_format is None:
            width_format = format

        self.proportion_adjusted_room_size = LinkedProportion(self, "room_size")
        if room_size_format == Constants.PERCENTAGE:
            self.proportion_adjusted_room_size.set_proportion_percentage(room_size)
        else:
            self.proportion_adjusted_room_size.set_proportion_decimal(room_size)

        self.proportion_adjusted_damping = LinkedProportion(self, "damping")
        if damping_format == Constants.PERCENTAGE:
            self.proportion_adjusted_damping.set_proportion_percentage(damping)
        else:
            self.proportion_adjusted_damping.set_proportion_decimal(damping)

        self.proportion_adjusted_wet_level = LinkedProportion(self, "wet_level")
        if wet_level_format == Constants.PERCENTAGE:
            self.proportion_adjusted_wet_level.set_proportion_percentage(wet_level)
        else:
            self.proportion_adjusted_wet_level.set_proportion_decimal(wet_level)

        self.proportion_adjusted_dry_level = LinkedProportion(self, "dry_level")
        if dry_level_format == Constants.PERCENTAGE:
            self.proportion_adjusted_dry_level.set_proportion_percentage(dry_level)
        else:
            self.proportion_adjusted_dry_level.set_proportion_decimal(dry_level)

        self.proportion_adjusted_width = LinkedProportion(self, "width")
        if width_format == Constants.PERCENTAGE:
            self.proportion_adjusted_width.set_proportion_percentage(width)
        else:
            self.proportion_adjusted_width.set_proportion_decimal(width)

        super().__init__(
            room_size=self.proportion_adjusted_room_size.get_proportion_decimal(),
            damping=self.proportion_adjusted_damping.get_proportion_decimal(),
            wet_level=self.proportion_adjusted_wet_level.get_proportion_decimal(),
            dry_level=self.proportion_adjusted_dry_level.get_proportion_decimal(),
            width=self.proportion_adjusted_width.get_proportion_decimal(),
            freeze_mode=freeze_mode)

    def set_freeze_mode(self, freeze_mode):
        """
        🟩 **R** -
        """
        self.freeze_mode = freeze_mode

    def get_freeze_mode(self):
        """
        🟩 **R** -
        """
        return self.freeze_mode


import threading
import subprocess

from pmma.core.py_src.Constants import Constants, InternalConstants

from pmma.build.General import General

class Executor:
    def __init__(self):
        self._exit_code = None
        self._result = None
        self._thread = None

    def run(self, command, blocking=True, hide_window=True):
        self._exit_code = None
        self._result = None

        self._thread = threading.Thread(target=self._run, args=(command, hide_window,))
        self._thread.name = "Executor:Execution_Thread"
        if blocking is False:
            self._thread.daemon = True
        self._thread.start()

        if blocking:
            self._thread.join()

    def _run(self, command, hide_window):
        command_type = type(command)

        try:
            if command_type == list or command_type == tuple:
                if hide_window and General.get_operating_system() == Constants.WINDOWS:
                    result = subprocess.run(command, capture_output=True, text=True, creationflags=InternalConstants.CREATE_NO_WINDOW)
                else:
                    result = subprocess.run(command, capture_output=True, text=True)
            else:
                print("You are not using an array of arguments as your command. \
This has the potential to be less secure, especially when using the user's input as a \
command. It is strongly recommended that you change your approach to use a list to avoid \
'shell injection vulnerabilities' where the end user specifies the command to be run, not \
its arguments, leading to unsecure commands being run on the host system!")

                if hide_window and General.get_operating_system() == Constants.WINDOWS:
                    result = subprocess.run(command, shell=True, capture_output=True, text=True, creationflags=InternalConstants.CREATE_NO_WINDOW)
                else:
                    result = subprocess.run(command, shell=True, capture_output=True, text=True)

            self._result = result.stdout
            self._exit_code = result.returncode
        except subprocess.CalledProcessError as result:
            self._result = result.output
            self._exit_code = result.returncode

    def get_exit_code(self):
        return self._exit_code

    def get_result(self):
        if type(self._result) == str:
            return self._result.strip()
        else:
            return self._result

class AdvancedExecutor:
    def __init__(self):
        self._exit_code = None
        self._result = ""
        self._thread = None

        self._command_running = False

    def run(self, command, hide_window=True):
        if self._command_running is False:
            self._command_running = True

            self._exit_code = None
            self._result = ""

            self._thread = threading.Thread(target=self._update_result, args=(command, hide_window,))
            self._thread.daemon = True
            self._thread.name = "AdvancedExecutor:Execution_Thread"
            self._thread.start()

    def get_busy(self):
        return self._command_running

    def get_result(self):
        return self._result.strip()

    def _update_result(self, command, hide_window):
        for path in self._run(command, hide_window):
            self._result += path.strip() + "\n"
        self._command_running = False

    def _run(self, command, hide_window):
        command_type = type(command)
        if command_type == list or command_type == tuple:
            if hide_window and General.get_operating_system() == Constants.WINDOWS:
                process = subprocess.Popen(command, stdout=subprocess.PIPE, text=True, creationflags=InternalConstants.CREATE_NO_WINDOW)
            else:
                process = subprocess.Popen(command, stdout=subprocess.PIPE, text=True)
        else:
            print("You are not using an array of arguments as your command. \
This has the potential to be less secure, especially when using the user's input as a \
command. It is strongly recommended that you change your approach to use a list to avoid \
'shell injection vulnerabilities' where the end user specifies the command to be run, not \
its arguments, leading to unsecure commands being run on the host system!")

            if hide_window and General.get_operating_system() == Constants.WINDOWS:
                process = subprocess.Popen(command, stdout=subprocess.PIPE, shell=True, text=True, creationflags=InternalConstants.CREATE_NO_WINDOW)
            else:
                process = subprocess.Popen(command, stdout=subprocess.PIPE, shell=True, text=True)

        result = ""
        while True:
            output = process.stdout.readline()
            output = output.strip()
            if output == '' and process.poll() is not None:
                self._exit_code = process.returncode
                break
            if output:
                result += output.strip()
                yield output.strip()
        rc = process.poll()
        yield str(rc)


from typing import Union, Iterable

import numpy as np
import numpy.typing as npt

Numerical1D = Union[
    npt.NDArray[np.float32], # preferred
    npt.NDArray[np.float16],
    npt.NDArray[np.float64],
    npt.NDArray[np.int32], # preferred
    npt.NDArray[np.int8],
    npt.NDArray[np.int16],
    npt.NDArray[np.int64],
    Iterable[float],  # preferred
    Iterable[int],  # preferred
]

Numerical = Union[float, int]

class TextRenderer:
    def set_text(self, text: str) -> None: ...
    def set_font(self, font: str) -> None: ...

    def set_size(self, size: Numerical) -> None: ...

    def set_foreground_color(self, color: Numerical1D) -> None: ...
    def set_background_color(self, color: Numerical1D) -> None: ...

    def set_position(self, position: Numerical1D) -> None: ...

    def render(self) -> None: ...


from typing import Union

Numerical = Union[float, int]

class ButtonPressedEvent:
    def get_pressed(self) -> bool: ...
    def get_pressed_toggle(self) -> bool: ...
    def get_double_pressed(self) -> bool: ...
    def get_long_pressed(self) -> bool: ...

    def poll_long_pressed(self) -> bool: ...

    def get_repeat_press_duration(self) -> Numerical: ...
    def get_long_press_duration(self) -> Numerical: ...
    def get_double_press_duration(self) -> Numerical: ...

    def set_repeat_press_duration(self, duration: Numerical) -> None: ...
    def set_double_press_duration(self, duration: Numerical) -> None: ...
    def set_long_press_duration(self, duration: Numerical) -> None: ...

class Zero(ButtonPressedEvent):
    def __init__(self) -> None: ...

class One(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Two(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Three(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Four(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Five(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Siz(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Seven(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Eight(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Nine(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Decimal(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Divide(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Multiply(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Subtract(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Add(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Enter(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Equal(ButtonPressedEvent):
    def __init__(self) -> None: ...


from tkinter import ttk

from pmma.core.py_src.Constants import InternalConstants

class Tkinter:
    @staticmethod
    def style(widget):
        if widget not in InternalConstants.TKINTER_STYLES:
            print("This is not a valid widget name, the valid widget names are: Constants.TKINTER_STYLE_BUTTON, Constants.TKINTER_STYLE_CHECKBUTTON, Constants.TKINTER_STYLE_COMBOBOX, Constants.TKINTER_STYLE_ENTRY, Constants.TKINTER_STYLE_FRAME, Constants.TKINTER_STYLE_LABEL, Constants.TKINTER_STYLE_LABELFRAME, Constants.TKINTER_STYLE_MENUBUTTON, Constants.TKINTER_STYLE_NOTEBOOK, Constants.TKINTER_STYLE_PANEDWINDOW, Constants.TKINTER_STYLE_HORIZONTAL_PROGRESSBAR, Constants.TKINTER_STYLE_VERTICAL_PROGRESSBAR, Constants.TKINTER_STYLE_RADIOBUTTON, Constants.TKINTER_STYLE_HORIZONTAL_SCALE, Constants.TKINTER_STYLE_VERTICAL_SCALE, Constants.TKINTER_STYLE_HORIZONTAL_SCROLLBAR, Constants.TKINTER_STYLE_VERTICAL_SCROLLBAR, Constants.TKINTER_STYLE_SEPARATOR, Constants.TKINTER_STYLE_SIZEGRIP and Constants.TKINTER_STYLE_TREEVIEW.")
            return

        style = ttk.Style()
        style.configure(
            widget,
            background="white",
            foreground="black",
            borderwidth=1,
            focusthickness=3,
            focuscolor="none")

        style.map(
            widget,
            background=[
                ("active", "white")])


class NoInputDevicesFoundError(Exception): ...

class UnableToReadAudioSampleError(Exception): ...

class PassportNotInitializedError(Exception): ...

class PassportNotRegisteredError(Exception): ...

class DirectoryNotFoundError(Exception): ...

class DirectoryAlreadyExistsError(Exception): ...



import numpy

class Stack:
    def __init__(self, max_size=None):
        self._frames = []
        self._max_size = max_size
        self._has_changed = False

    def push(self, item):
        if not self.is_full():
            self._has_changed = True
            self._frames.append(item)

    def pop(self):
        if not self.is_empty():
            self.has_changed = True
            return self._frames.pop()

    def peek(self):
        if not self.is_empty():
            return self._frames[-1]

    def is_empty(self):
        return len(self._frames) == 0

    def is_full(self):
        if self._max_size is None:
            return False
        return len(self._frames) >= self._max_size

    def size(self):
        return len(self._frames)

    def clear(self):
        self._has_changed = True
        self._frames = []

    def changed(self):
        value = self._has_changed
        self._has_changed = False
        return value

class Queue:
    def __init__(self, max_size=None):
        self._frames = []
        self._max_size = max_size
        self.has_changed = False

    def enqueue(self, item):
        if not self.is_full():
            self._has_changed = True
            self._frames.append(item)

    def dequeue(self):
        if not self.is_empty():
            self._has_changed = True
            return self._frames[0]

    def peek(self):
        if not self.is_empty():
            return self._frames[0]

    def is_empty(self):
        return len(self._frames) == 0

    def is_full(self):
        if self._max_size is None:
            return False
        return len(self._frames) >= self._max_size

    def size(self):
        return len(self._frames)

    def clear(self):
        self._has_changed = True
        self._frames = []

    def changed(self):
        value = self._has_changed
        self._has_changed = False
        return value

class CircularQueue:
    def __init__(self, size):
        self._max_size = size
        self._frames = [0] * size
        self._front = -1
        self._rear = -1
        self._has_changed = False

    def clear(self):
        self._has_changed = True
        self._frames = [0] * self._max_size
        self._front = -1
        self._rear = -1

    def enqueue(self, item):
        if self.is_empty():
            self._front = 0
            self._rear = 0
            self._frames[self._rear] = item
        else:
            self._has_changed = True
            self._rear = (self._rear + 1) % self._max_size
            if self.is_full():
                self._rear = (self._rear - 1 + self._max_size) % self._max_size
            else:
                self._frames[self._rear] = item

    def dequeue(self):
        if not self.is_empty():
            self._has_changed = True
            item = self._frames[self._front]
            if self._front == self._rear:
                self._front = -1
                self._rear = -1
            else:
                self._front = (self._front + 1) % self._max_size

            return item

    def peek(self):
        if not self.is_empty():
            return self._frames[self._front]

    def size(self):
        if self.is_empty():
            return 0
        elif self._front <= self._rear:
            return self._rear - self._front + 1
        else:
            return self._max_size - self._front + self._rear + 1

    def is_empty(self):
        return self._front == -1 and self._rear == -1

    def is_full(self):
        rear = (self._rear + 1) % self._max_size
        return rear == self._front

    def changed(self):
        value = self._has_changed
        self._has_changed = False
        return value

class PriorityQueue:
    def __init__(self):
        # Initialize an empty list to store the heap as an array of tuples (priority, value)
        self._heap = numpy.array([], dtype=[('priority', numpy.int32), ('value', object)])
        self._has_changed = False

    def enqueue(self, item, priority):
        # Append the new (priority, value) tuple to the heap
        new_item = numpy.array([(priority, item)], dtype=self._heap.dtype)
        self._heap = numpy.append(self._heap, new_item)
        # Perform up-heap bubbling to maintain the max-heap property based on priority
        self._sift_up(len(self._heap) - 1)
        self._has_changed = True

    def dequeue(self):
        if len(self._heap) != 0:
            self._has_changed = True
            # The root of the heap (index 0) has the maximum priority
            max_value = self._heap[0]['value']

            # Move the last element to the root and then heapify down
            self._heap[0] = self._heap[-1]
            self._heap = numpy.delete(self._heap, -1)

            # Perform down-heap bubbling to maintain the max-heap property
            self._sift_down(0)

            return max_value

    def peek_next_priority(self):
        if len(self._heap) != 0:
            return self._heap[0]['priority']

    def peek(self):
        if len(self._heap) != 0:
            return self._heap[0]['value']

    def _sift_up(self, index):
        parent = (index - 1) // 2
        if index > 0 and self._heap[index]['priority'] > self._heap[parent]['priority']:
            self._heap[[index, parent]] = self._heap[[parent, index]]
            self._sift_up(parent)

    def _sift_down(self, index):
        largest = index
        left = 2 * index + 1
        right = 2 * index + 2

        if left < len(self._heap) and self._heap[left]['priority'] > self._heap[largest]['priority']:
            largest = left
        if right < len(self._heap) and self._heap[right]['priority'] > self._heap[largest]['priority']:
            largest = right

        if largest != index:
            self._heap[[index, largest]] = self._heap[[largest, index]]
            self._sift_down(largest)

    def is_empty(self):
        return len(self._heap) == 0

    def size(self):
        return len(self._heap)

    def clear(self):
        self._heap = numpy.array([], dtype=self._heap.dtype)
        self._has_changed = True

    def changed(self):
        value = self._has_changed
        self._has_changed = False
        return value

class InvertedPriorityQueue:
    def __init__(self):
        # Initialize an empty list to store the heap as an array of tuples (priority, value)
        self._heap = numpy.array([], dtype=[('priority', numpy.int32), ('value', object)])
        self._has_changed = False

    def enqueue(self, item, priority):
        # Append the new (priority, value) tuple to the heap
        new_item = numpy.array([(priority, item)], dtype=self._heap.dtype)
        self._heap = numpy.append(self._heap, new_item)
        # Perform up-heap bubbling to maintain the min-heap property based on priority
        self._sift_up(len(self._heap) - 1)
        self._has_changed = True

    def dequeue(self):
        if len(self._heap) != 0:
            self._has_changed = True
            # The root of the heap (index 0) has the minimum priority
            min_value = self._heap[0]['value']

            # Move the last element to the root and then heapify down
            self._heap[0] = self._heap[-1]
            self._heap = numpy.delete(self._heap, -1)

            # Perform down-heap bubbling to maintain the min-heap property
            self._sift_down(0)

            return min_value

    def peek_next_priority(self):
        if len(self._heap) != 0:
            return self._heap[0]['priority']

    def peek(self):
        if len(self._heap) != 0:
            return self._heap[0]['value']

    def _sift_up(self, index):
        parent = (index - 1) // 2
        if index > 0 and self._heap[index]['priority'] < self._heap[parent]['priority']:
            self._heap[[index, parent]] = self._heap[[parent, index]]
            self._sift_up(parent)

    def _sift_down(self, index):
        smallest = index
        left = 2 * index + 1
        right = 2 * index + 2

        if left < len(self._heap) and self._heap[left]['priority'] < self._heap[smallest]['priority']:
            smallest = left
        if right < len(self._heap) and self._heap[right]['priority'] < self._heap[smallest]['priority']:
            smallest = right

        if smallest != index:
            self._heap[[index, smallest]] = self._heap[[smallest, index]]
            self._sift_down(smallest)

    def is_empty(self):
        return len(self._heap) == 0

    def size(self):
        return len(self._heap)

    def clear(self):
        self._heap = numpy.array([], dtype=self._heap.dtype)
        self._has_changed = True

    def changed(self):
        value = self._has_changed
        self._has_changed = False
        return value

class PriorityList:
    def __init__(self):
        # Initialize an empty list to store the heap as an array of tuples (priority, value)
        self._heap = numpy.array([], dtype=[('priority', numpy.int32), ('value', object)])
        self._has_changed = False

    def add(self, item, priority):
        # Append the new (priority, value) tuple to the heap
        new_item = numpy.array([(priority, item)], dtype=self._heap.dtype)
        self._heap = numpy.append(self._heap, new_item)
        # Perform up-heap bubbling to maintain the max-heap property based on priority
        self._sift_up(len(self._heap) - 1)
        self._has_changed = True

    def remove_item(self, item):
        for i in range(len(self._heap)):
            if self._heap[i]['value'] == item:
                self._heap = numpy.delete(self._heap, i)
                self._has_changed = True
                break

    def remove_highest_priority(self):
        values = []
        priority = self.peek_next_priority()
        while not self.is_empty() and self.peek_next_priority() == priority:
            # The root of the heap (index 0) has the maximum priority
            max_value = self._heap[0]['value']

            # Move the last element to the root and then heapify down
            self._heap[0] = self._heap[-1]
            self._heap = numpy.delete(self._heap, -1)

            # Perform down-heap bubbling to maintain the max-heap property
            self._sift_down(0)

            values.append(max_value)
            self._has_changed = True
        return values

    def update_priority(self, item, new_priority):
        for i in range(len(self._heap)):
            if self._heap[i]['value'] == item:
                self._heap[i]['priority'] = new_priority
                self._sift_up(i)
                self._sift_down(i)
                self._has_changed = True
                break

    def peek_next_priority(self):
        if len(self._heap) != 0:
            return self._heap[0]['priority']

    def peek(self):
        if len(self._heap) != 0:
            return self._heap[0]['value']

    def _sift_up(self, index):
        parent = (index - 1) // 2
        if index > 0 and self._heap[index]['priority'] > self._heap[parent]['priority']:
            self._heap[[index, parent]] = self._heap[[parent, index]]
            self._sift_up(parent)

    def _sift_down(self, index):
        largest = index
        left = 2 * index + 1
        right = 2 * index + 2

        if left < len(self._heap) and self._heap[left]['priority'] > self._heap[largest]['priority']:
            largest = left
        if right < len(self._heap) and self._heap[right]['priority'] > self._heap[largest]['priority']:
            largest = right

        if largest != index:
            self._heap[[index, largest]] = self._heap[[largest, index]]
            self._sift_down(largest)

    def is_empty(self):
        return len(self._heap) == 0

    def size(self):
        return len(self._heap)

    def clear(self):
        self._heap = numpy.array([], dtype=self._heap.dtype)
        self._has_changed = True

    def changed(self):
        value = self._has_changed
        self._has_changed = False
        return value

class InvertedPriorityList:
    def __init__(self):
        # Initialize an empty list to store the heap as an array of tuples (priority, value)
        self._heap = numpy.array([], dtype=[('priority', numpy.int32), ('value', object)])
        self._has_changed = False

    def add(self, item, priority):
        # Append the new (priority, value) tuple to the heap
        new_item = numpy.array([(priority, item)], dtype=self._heap.dtype)
        self._heap = numpy.append(self._heap, new_item)
        # Perform up-heap bubbling to maintain the min-heap property based on priority
        self._sift_up(len(self._heap) - 1)
        self._has_changed = True

    def remove_highest_priority(self):
        values = []
        priority = self.peek_next_priority()
        while not self.is_empty() and self.peek_next_priority() == priority:
            # The root of the heap (index 0) has the minimum priority
            min_value = self._heap[0]['value']

            # Move the last element to the root and then heapify down
            self._heap[0] = self._heap[-1]
            self._heap = numpy.delete(self._heap, -1)

            # Perform down-heap bubbling to maintain the min-heap property
            self._sift_down(0)

            values.append(min_value)
            self._has_changed = True

        return values

    def update_priority(self, item, new_priority):
        for i in range(len(self._heap)):
            if self._heap[i]['value'] == item:
                self._heap[i]['priority'] = new_priority
                self._sift_up(i)
                self._sift_down(i)
                self._has_changed = True
                break

    def remove_item(self, item):
        for i in range(len(self._heap)):
            if self._heap[i]['value'] == item:
                self._heap = numpy.delete(self._heap, i)
                self._has_changed = True
                break

    def peek_next_priority(self):
        if len(self._heap) != 0:
            return self._heap[0]['priority']

    def peek(self):
        if len(self._heap) != 0:
            return self._heap[0]['value']

    def _sift_up(self, index):
        parent = (index - 1) // 2
        if index > 0 and self._heap[index]['priority'] < self._heap[parent]['priority']:
            self._heap[[index, parent]] = self._heap[[parent, index]]
            self._sift_up(parent)

    def _sift_down(self, index):
        smallest = index
        left = 2 * index + 1
        right = 2 * index + 2

        if left < len(self._heap) and self._heap[left]['priority'] < self._heap[smallest]['priority']:
            smallest = left
        if right < len(self._heap) and self._heap[right]['priority'] < self._heap[smallest]['priority']:
            smallest = right

        if smallest != index:
            self._heap[[index, smallest]] = self._heap[[smallest, index]]
            self._sift_down(smallest)

    def is_empty(self):
        return len(self._heap) == 0

    def size(self):
        return len(self._heap)

    def clear(self):
        self._heap = numpy.array([], dtype=self._heap.dtype)
        self._has_changed = True

    def changed(self):
        value = self._has_changed
        self._has_changed = False
        return value


from typing import Iterable, Union

import numpy as np
import numpy.typing as npt

Numerical = Union[float, int]

Numerical1D = Union[
    npt.NDArray[np.float32], # preferred
    npt.NDArray[np.float16],
    npt.NDArray[np.float64],
    npt.NDArray[np.int32], # preferred
    npt.NDArray[np.int8],
    npt.NDArray[np.int16],
    npt.NDArray[np.int64],
    Iterable[float],  # preferred
    Iterable[int],  # preferred
]

class ButtonPressedEvent:
    def get_pressed(self) -> bool: ...
    def get_pressed_toggle(self) -> bool: ...
    def get_double_pressed(self) -> bool: ...
    def get_long_pressed(self) -> bool: ...

    def poll_long_pressed(self) -> bool: ...

    def get_repeat_press_duration(self) -> Numerical: ...
    def get_long_press_duration(self) -> Numerical: ...
    def get_double_press_duration(self) -> Numerical: ...

    def set_repeat_press_duration(self, duration: Numerical) -> None: ...
    def set_double_press_duration(self, duration: Numerical) -> None: ...
    def set_long_press_duration(self, duration: Numerical) -> None: ...

class MousePosition:
    def __init__(self) -> None: ...

    def get_position(self, using_numpy: bool=True) -> Numerical1D: ...

    def get_delta(self, using_numpy: bool=True) -> Numerical1D: ...
    def get_delta_toggle(self, using_numpy: bool=True) -> Numerical1D: ...

class MouseEnterWindow:
    def __init__(self) -> None: ...

    def get_entered(self) -> bool: ...
    def get_entered_toggle(self) -> bool: ...

class MouseButton_Left(ButtonPressedEvent):
    def __init__(self) -> None: ...

class MouseButton_Right(ButtonPressedEvent):
    def __init__(self) -> None: ...

class MouseButton_Middle(ButtonPressedEvent):
    def __init__(self) -> None: ...

class MouseButton_0(ButtonPressedEvent):
    def __init__(self) -> None: ...

class MouseButton_1(ButtonPressedEvent):
    def __init__(self) -> None: ...

class MouseButton_2(ButtonPressedEvent):
    def __init__(self) -> None: ...

class MouseButton_3(ButtonPressedEvent):
    def __init__(self) -> None: ...

class MouseButton_4(ButtonPressedEvent):
    def __init__(self) -> None: ...

class MouseScroll:
    def __init__(self) -> None: ...

    def get_position(self, using_numpy: bool=True) -> Numerical1D: ...
    def clear_position(self) -> None: ...
    def get_horizontal_position(self) -> Numerical: ...
    def get_vertical_position(self) -> Numerical: ...

    def get_delta(self, using_numpy: bool=True) -> Numerical1D: ...
    def get_horizontal_delta(self) -> Numerical: ...
    def get_vertical_delta(self) -> Numerical: ...

    def get_delta_toggle(self, using_numpy: bool=True) -> Numerical1D: ...
    def get_horizontal_delta_toggle(self) -> Numerical: ...
    def get_vertical_delta_toggle(self) -> Numerical: ...

    def get_enabled(self) -> bool: ...

    def set_enabled(self, value: bool) -> None: ...


class File:
    def __init__(self, file_path: str) -> None: ...
    def get_absolute_file_path(self) -> None: ...
    def get_relative_file_path(self) -> None: ...
    def is_file_exists(self) -> None: ...
    def move(self, destination: str) -> None: ...
    def copy(self, destination: str) -> None: ...
    def delete(self) -> None: ...
    def recycle(self) -> None: ...
    def rename(self, new_name: str) -> None: ...

class Directory:
    def __init__(self, directory_name: str) -> None: ...
    def get_absolute_path(self) -> None: ...
    def get_relative_path(self) -> None: ...
    def is_directory_exists(self) -> None: ...
    def move(self, destination: str) -> None: ...
    def copy(self, destination: str) -> None: ...
    def delete(self) -> None: ...
    def recycle(self) -> None: ...
    def rename(self, new_name: str) -> None: ...


from pmma.build.NumberFormats import DisplayCoordinate

class Linear:
    start_position: DisplayCoordinate
    end_position: DisplayCoordinate

    def __init__(self, DisplayCoordinateInstance: DisplayCoordinate) -> None: ...
    def start(self) -> None: ...
    def stop(self) -> None: ...
    def pause(self) -> None: ...
    def resume(self) -> None: ...
    def set_duration(self, duration: float) -> None: ...
    def set_looping(self, looping: bool) -> None: ...
    def set_repeating(self, repeating: bool) -> None: ...

    def get_duration(self) -> float: ...
    def get_remaining_duration(self) -> float: ...

    def is_paused(self) -> bool: ...
    def is_playing(self) -> bool: ...
    def is_repeating(self) -> bool: ...
    def is_looping(self) -> bool: ...

class Radial:
    start_position: DisplayCoordinate
    orbit_position: DisplayCoordinate

    def __init__(self, DisplayCoordinateInstance: DisplayCoordinate) -> None: ...
    def start(self) -> None: ...
    def stop(self) -> None: ...
    def pause(self) -> None: ...
    def resume(self) -> None: ...
    def set_duration(self, duration: float) -> None: ...
    def set_repeating(self, repeating: bool) -> None: ...

    def get_duration(self) -> float: ...
    def get_remaining_duration(self) -> float: ...

    def is_paused(self) -> bool: ...
    def is_playing(self) -> bool: ...
    def is_repeating(self) -> bool: ...


class Constants:
    HAT_NOT_PRESSED = "HAT NOT PRESSED"
    HAT_PRESSED_UP = "HAT PRESSED UP"
    HAT_PRESSED_RIGHT = "HAT PRESSED RIGHT"
    HAT_PRESSED_DOWN = "HAT PRESSED DOWN"
    HAT_PRESSED_LEFT = "HAT PRESSED LEFT"
    HAT_PRESSED_UP_RIGHT = "HAT PRESSED UP-RIGHT"
    HAT_PRESSED_DOWN_RIGHT = "HAT PRESSED DOWN-RIGHT"
    HAT_PRESSED_DOWN_LEFT = "HAT PRESSED DOWN-LEFT"
    HAT_PRESSED_UP_LEFT = "HAT PRESSED UP-LEFT"

    DECIMAL = "decimal"
    PERCENTAGE = "percentage"

    VALUE = "value"
    UPDATING = "updating"
    MANUALLY_SET = "manually_set"

    WINDOWS = "windows"
    LINUX = "linux"
    MACOS = "macOS"
    JAVA = "java"
    ANDROID = "android"

    TKINTER_STYLE_BUTTON = "TButton"
    TKINTER_STYLE_CHECKBUTTON = "TCheckbutton"
    TKINTER_STYLE_COMBOBOX = "TCombobox"
    TKINTER_STYLE_ENTRY = "TEntry"
    TKINTER_STYLE_FRAME = "TFrame"
    TKINTER_STYLE_LABEL = "TLabel"
    TKINTER_STYLE_LABELFRAME = "TLabelFrame"
    TKINTER_STYLE_MENUBUTTON = "TMenubutton"
    TKINTER_STYLE_NOTEBOOK = "TNotebook"
    TKINTER_STYLE_PANEDWINDOW = "TPanedwindow"
    TKINTER_STYLE_HORIZONTAL_PROGRESSBAR = "Horizontal.TProgressbar"
    TKINTER_STYLE_VERTICAL_PROGRESSBAR = "Vertical.TProgressbar"
    TKINTER_STYLE_RADIOBUTTON = "TRadiobutton"
    TKINTER_STYLE_HORIZONTAL_SCALE = "Horizontal.TScale"
    TKINTER_STYLE_VERTICAL_SCALE = "Vertical.TScale"
    TKINTER_STYLE_HORIZONTAL_SCROLLBAR = "Horizontal.TScrollbar"
    TKINTER_STYLE_VERTICAL_SCROLLBAR = "Vertical.TScrollbar"
    TKINTER_STYLE_SEPARATOR = "TSeparator"
    TKINTER_STYLE_SIZEGRIP = "TSizegrip"
    TKINTER_STYLE_TREEVIEW = "Treeview"

class InternalConstants:
    DATA_COLLECTION_METHODS = "data_collection_methods"

    SMI = "smi"
    WMI = "wmi"
    PYADL = "pyadl"

    CREATE_NO_WINDOW = 0x08000000

    TKINTER_STYLES = [
        Constants.TKINTER_STYLE_BUTTON,
        Constants.TKINTER_STYLE_CHECKBUTTON,
        Constants.TKINTER_STYLE_COMBOBOX,
        Constants.TKINTER_STYLE_ENTRY,
        Constants.TKINTER_STYLE_FRAME,
        Constants.TKINTER_STYLE_LABEL,
        Constants.TKINTER_STYLE_LABELFRAME,
        Constants.TKINTER_STYLE_MENUBUTTON,
        Constants.TKINTER_STYLE_NOTEBOOK,
        Constants.TKINTER_STYLE_PANEDWINDOW,
        Constants.TKINTER_STYLE_HORIZONTAL_PROGRESSBAR,
        Constants.TKINTER_STYLE_VERTICAL_PROGRESSBAR,
        Constants.TKINTER_STYLE_RADIOBUTTON,
        Constants.TKINTER_STYLE_HORIZONTAL_SCALE,
        Constants.TKINTER_STYLE_VERTICAL_SCALE,
        Constants.TKINTER_STYLE_HORIZONTAL_SCROLLBAR,
        Constants.TKINTER_STYLE_VERTICAL_SCROLLBAR,
        Constants.TKINTER_STYLE_SEPARATOR,
        Constants.TKINTER_STYLE_SIZEGRIP,
        Constants.TKINTER_STYLE_TREEVIEW
    ]

    SECONDS_PER_MINUTE = 60
    SECONDS_PER_HOUR = SECONDS_PER_MINUTE * 60
    SECONDS_PER_DAY = SECONDS_PER_HOUR * 24
    DAYS_PER_YEAR = 365.25  # accounting for leap years
    DAYS_PER_MONTH = DAYS_PER_YEAR / 12
    HOURS_PER_DAY = 24
    MINUTES_PER_HOUR = 60
    MICROSECONDS_PER_SECOND = 1e6


from typing import Union, Iterable, Any

class Executor:
    def __init__(self) -> None: ...

    def run(self, command: Union[str, Iterable[str]], blocking: bool=True, hide_window:bool=True) -> None: ...

    def get_exit_code(self) -> int: ...

    def get_result(self) -> Any: ...

class AdvancedExecutor:
    def __init__(self) -> None: ...

    def run(self, command: Union[str, Iterable[str]], hide_window: bool=True) -> None: ...

    def get_busy(self) -> bool: ...

    def get_result(self) -> Any: ...


import threading, time

import pyaudio, numpy

import pmma.core.py_src.Error as Error
from pmma.build.Logger import Logger

class Sampler:
    def __init__(self, chunk_size=2048, sampling_rate=44100, input_device_id=None):
        self.logger = Logger()

        self._pyaudio_instance = pyaudio.PyAudio()

        if input_device_id is None:
            self._input_device = self.get_default_input_device()
        else:
            self._input_device = input_device_id

        self._do_sampling = True
        self._is_sampling_running = False
        self._do_pause_sampling = False

        self._sampling_rate = sampling_rate
        self._chunk_size = chunk_size

        self._volume = 0
        self._frequency = []
        self._loudest_frequency = 0

        self._sampler_thread = threading.Thread(
            target=self.sampler)
        self._sampler_thread.daemon = True
        self._sampler_thread.name = "Sampler:Sampler_Thread"

    def __del__(self):
        self._do_sampling = False

        if self._is_sampling_running:
            self._sampler_thread.join()

    def print_input_devices(self):
        info = self._pyaudio_instance.get_host_api_info_by_index(0)
        numdevices = info.get('deviceCount')
        for i in range(numdevices):
            if (self._pyaudio_instance.get_device_info_by_host_api_device_index(0, i).get('maxInputChannels')) > 0:
                input_device_name = self._pyaudio_instance.get_device_info_by_host_api_device_index(0, i).get('name')
                print(f"Input Device id {i} - {input_device_name}")

    def get_input_devices(self):
        input_devices = []
        info = self._pyaudio_instance.get_host_api_info_by_index(0)
        numdevices = info.get('deviceCount')
        for i in range(numdevices):
            if (self._pyaudio_instance.get_device_info_by_host_api_device_index(0, i).get('maxInputChannels')) > 0:
                input_device_name = self._pyaudio_instance.get_device_info_by_host_api_device_index(0, i).get('name')
                input_devices.append(input_device_name)
        return input_devices

    def set_input_device(self, input_device_id=None):
        if input_device_id is None:
            input_device = self.get_default_input_device()
        else:
            input_device = input_device_id

        if self._input_device != input_device:
            self._input_device = input_device

            if self._is_sampling_running:
                self.stop()
                self.start()

    def get_default_input_device(self):
        try:
            return self._pyaudio_instance.get_default_input_device_info()["index"]
        except IOError:
            return

    def sampler(self, wait_time):
        self._is_sampling_running = True

        try:
            if self._input_device is None:
                self.logger.internal_log_error(
                    15,
                    "No input device was found. The easiest reason \
for this is because the operating system hasn't detected any input device \
either. Therefore the first step for troubleshooting this is to make sure \
that you have an input device connected to your current machine. Likewise \
also make sure that the device is behaving as expected. Should this fail and \
the operating system has detected your device, consider running the \
'print_input_devices()' command and manually selecting the device you \
want to use.",
                    True)

                raise Error.NoInputDevicesFoundError("No audio input devices were found!")

            stream = self._pyaudio_instance.open(
                format=pyaudio.paInt16,
                channels=1,
                rate=self._sampling_rate,
                input=True,
                frames_per_buffer=self._chunk_size,
                input_device_index=self._input_device)

            while self._do_sampling:
                if self._do_pause_sampling:
                    if wait_time is None:
                        wait_time = 1/60
                    time.sleep(wait_time)
                    continue

                try:
                    stream_data = stream.read(self._chunk_size)
                except Exception as error:
                    self.logger.internal_log_error(
                        16,
                        "Unable to read the current audio sample. Whilst \
the exact cause of this error is unknown, it's likely that the cause for \
this error was that the audio input device was disconnected without first \
having had the sampling process stopped. Please use the existing 'stop()' \
method in order to avoid this error being raised when the audio device \
is removed.",
                        True)
                    raise Error.UnableToReadAudioSampleError("Unable to read audio sample!") from error

                data = numpy.frombuffer(
                    stream_data,
                    dtype=numpy.int16)

                peak = numpy.average(numpy.abs(data))*2

                self._volume = peak

                data = data * numpy.hanning(len(data))
                self._frequency = abs(numpy.fft.fft(data).real)

                fft = self._frequency[:int(len(self._frequency)/2)]
                freq = numpy.fft.fftfreq(self._chunk_size, 1.0/self._sampling_rate)
                freq = freq[:int(len(freq)/2)]
                self._loudest_frequency = freq[numpy.where(fft==numpy.max(fft))[0][0]]+1

            self._is_sampling_running = False
            stream.stop_stream()
            stream.close()
        except Exception as error:
            self._is_sampling_running = False
            raise error

    def start(self, wait_time=None):
        if self._is_sampling_running is False:
            self._do_sampling = True
            self._is_sampling_running = True

            self._sampler_thread = threading.Thread(
                target=self.sampler, args=[wait_time])
            self._sampler_thread.daemon = True
            self._sampler_thread.name = "Sampler:Sampler_Thread"
            self._sampler_thread.start()

    def stop(self, wait_until_stopped=True):
        self._do_sampling = False
        if wait_until_stopped:
            self._sampler_thread.join()

    def pause(self):
        self._do_pause_sampling = True

    def unpause(self):
        self._do_pause_sampling = False

    def get_volume(self):
        return self._volume

    def get_frequency(self):
        return self._frequency

    def get_loudest_frequency(self):
        return self._loudest_frequency

    def is_sampling(self):
        return self._is_sampling_running


from typing import Literal, Union, Tuple, Any, Callable

NoneFloat = Union[float, None]
NoneString = Union[str, None]

class General:
    @staticmethod
    def is_power_saving_mode_enabled() -> bool: ...

    @staticmethod
    def get_operating_system() -> Literal["windows", "linux", "macOS", "java", "android"]: ...

    @staticmethod
    def find_executable_nvidia_smi() -> str: ...

    @staticmethod
    def get_pmma_install_dir() -> str: ...

    @staticmethod
    def set_debug_mode(is_debug_mode: bool) -> None: ...

    @staticmethod
    def is_debug_mode_enabled() -> bool: ...

    @staticmethod
    def is_window_created() -> bool: ...

    @staticmethod
    def is_escape_key_to_close_window() -> bool: ...

    @staticmethod
    def set_escape_key_to_close_window(escape_key_to_close_window: bool) -> None: ...

    @staticmethod
    def is_f11_key_to_toggle_full_screen() -> bool: ...

    @staticmethod
    def set_f11_key_to_toggle_full_screen(f11_key_to_toggle_full_screen: bool) -> None: ...

    @staticmethod
    def is_application_running() -> None: ...

    @staticmethod
    def get_current_pmma_version() -> str: ...

    @staticmethod
    def get_latest_pmma_version() -> str: ...

    @staticmethod
    def is_update_available() -> bool: ...

    @staticmethod
    def get_application_start_time() -> float: ...

    @staticmethod
    def get_application_run_time() -> float: ...

    @staticmethod
    def get_shape_quality() -> float: ...

    @staticmethod
    def set_shape_quality(value: NoneFloat=None) -> None: ...

    @staticmethod
    def get_execution_time(function: Callable[..., Any], *args: Any, **kwargs: Any) -> Tuple[float, Any]: ...

    @staticmethod
    def get_execution_inverse_time(function: Callable[..., Any], *args: Any, **kwargs: Any) -> Tuple[float, Any]: ...

    @staticmethod
    def set_locale(locale: NoneString=None) -> None: ...

    @staticmethod
    def get_locale() -> str: ...

    @staticmethod
    def profile_start() -> None: ...

    @staticmethod
    def profile_end() -> None: ...


from typing import Iterable, Union, Any

import numpy as np
import numpy.typing as npt

NPInteger1D = Union[
    npt.NDArray[np.int32], # preferred
    npt.NDArray[np.int8],
    npt.NDArray[np.int16],
    npt.NDArray[np.int64],
    Iterable[int],
]

NPFloat1D = Union[
    npt.NDArray[np.float32], # preferred
    npt.NDArray[np.float16],
    npt.NDArray[np.float64],
    Iterable[float]]

Integer1D = Union[
    Iterable[int],
]

Float1D = Union[
    Iterable[float]]

Numerical = Union[float, int]

NoneInt = Union[int, None]

class Color:
    def __init__(self) -> None: ...

    def configure(self, seed: NoneInt=None, octaves: int=2, lacunarity: float=0.75, gain: float=1.0) -> None: ...

    def get_seed(self) -> int: ...
    def get_octaves(self) -> int: ...
    def get_lacunarity(self) -> Numerical: ...
    def get_gain(self) -> Numerical: ...

    def get_set(self) -> bool: ...

    def generate_from_random(self, generate_alpha: bool=True) -> None: ...
    def generate_from_1D_perlin_noise(self, in_value: Numerical, generate_alpha: bool=True) -> None: ...
    def generate_from_2D_perlin_noise(self, in_value_one: Numerical, in_value_two: Numerical, generate_alpha: bool=True) -> None: ...
    def generate_from_3D_perlin_noise(self, in_value_one: Numerical, in_value_two: Numerical, in_value_three: Numerical, generate_alpha: bool=True) -> None: ...
    def generate_from_1D_fractal_brownian_motion(self, in_value: Numerical, generate_alpha: bool=True) -> None: ...
    def generate_from_2D_fractal_brownian_motion(self, in_value_one: Numerical, in_value_two: Numerical, generate_alpha: bool=True) -> None: ...
    def generate_from_3D_fractal_brownian_motion(self, in_value_one: Numerical, in_value_two: Numerical, in_value_three: Numerical, generate_alpha: bool=True) -> None: ...

    def set_RGBA_array(self, in_color: NPInteger1D) -> None: ...
    def set_rgba_array(self, in_color: NPFloat1D) -> None: ...
    def set_RGB_array(self, in_color: NPInteger1D) -> None: ...
    def set_rgb_array(self, in_color: NPFloat1D) -> None: ...
    def get_RGBA_array(self, detect_format: bool=True) -> NPInteger1D: ...
    def get_rgba_array(self, detect_format: bool=True) -> NPFloat1D: ...
    def get_RGB_array(self, detect_format: bool=True) -> NPInteger1D: ...
    def get_rgb_array(self, detect_format: bool=True) -> NPFloat1D: ...

    def set_RGBA(self, r: int, g: int, b: int, a: int) -> None: ...
    def set_rgba(self, r: float, g: float, b: float, a: float) -> None: ...
    def set_RGB(self, r: int, g: int, b: int, a: int) -> None: ...
    def set_rgb(self, r: float, g: float, b: float) -> None: ...
    def get_RGBA(self) -> Integer1D: ...
    def get_rgba(self) -> Float1D: ...
    def get_RGB(self) -> Integer1D: ...
    def get_rgb(self) -> Float1D: ...

class DisplayCoordinate:
    def __init__(self) -> None: ...

    def configure(self, seed: NoneInt=None, octaves: int=2, lacunarity: float=0.75, gain: float=1.0) -> None: ...

    def get_seed(self) -> int: ...
    def get_octaves(self) -> int: ...
    def get_lacunarity(self) -> Numerical: ...
    def get_gain(self) -> Numerical: ...
    def get_set(self) -> bool: ...

    def generate_from_random(self) -> None: ...
    def generate_from_1D_perlin_noise(self, in_value: Numerical) -> None: ...
    def generate_from_2D_perlin_noise(self, in_value_one: Numerical, in_value_two: Numerical) -> None: ...
    def generate_from_3D_perlin_noise(self, in_value_one: Numerical, in_value_two: Numerical, in_value_three: Numerical) -> None: ...
    def generate_from_1D_fractal_brownian_motion(self, in_value: Numerical) -> None: ...
    def generate_from_2D_fractal_brownian_motion(self, in_value_one: Numerical, in_value_two: Numerical) -> None: ...
    def generate_from_3D_fractal_brownian_motion(self, in_value_one: Numerical, in_value_two: Numerical, in_value_three: Numerical) -> None: ...

    def set_coord_array(self, in_coord: NPInteger1D) -> None: ...
    def get_coord_array(self, detect_format: bool=True) -> NPFloat1D: ...

    def set_coord(self, x: int, y: int) -> None: ...
    def get_coord(self) -> Float1D: ...

class Angle:
    def __init__(self) -> None: ...

    def configure(self, seed: NoneInt=None, octaves: int=2, lacunarity: float=0.75, gain: float=1.0) -> None: ...

    def get_seed(self) -> int: ...
    def get_octaves(self) -> int: ...
    def get_lacunarity(self) -> Numerical: ...
    def get_gain(self) -> Numerical: ...
    def get_set(self) -> bool: ...

    def generate_from_random(self) -> None: ...
    def generate_from_1D_perlin_noise(self, in_value: Numerical) -> None: ...
    def generate_from_2D_perlin_noise(self, in_value_one: Numerical, in_value_two: Numerical) -> None: ...
    def generate_from_3D_perlin_noise(self, in_value_one: Numerical, in_value_two: Numerical, in_value_three: Numerical) -> None: ...
    def generate_from_1D_fractal_brownian_motion(self, in_value: Numerical) -> None: ...
    def generate_from_2D_fractal_brownian_motion(self, in_value_one: Numerical, in_value_two: Numerical) -> None: ...
    def generate_from_3D_fractal_brownian_motion(self, in_value_one: Numerical, in_value_two: Numerical, in_value_three: Numerical) -> None: ...

    def set_degrees(self, value: Numerical) -> None: ...
    def set_radians(self, value: Numerical) -> None: ...

    def get_degrees(self) -> Numerical: ...
    def get_radians(self) -> Numerical: ...

class Proportion:
    def __init__(self) -> None: ...

    def configure(self, seed: NoneInt=None, octaves: int=2, lacunarity: float=0.75, gain: float=1.0) -> None: ...

    def get_seed(self) -> int: ...
    def get_octaves(self) -> int: ...
    def get_lacunarity(self) -> Numerical: ...
    def get_gain(self) -> Numerical: ...
    def get_set(self) -> bool: ...

    def generate_from_random(self) -> None: ...
    def generate_from_1D_perlin_noise(self, in_value: Numerical) -> None: ...
    def generate_from_2D_perlin_noise(self, in_value_one: Numerical, in_value_two: Numerical) -> None: ...
    def generate_from_3D_perlin_noise(self, in_value_one: Numerical, in_value_two: Numerical, in_value_three: Numerical) -> None: ...
    def generate_from_1D_fractal_brownian_motion(self, in_value: Numerical) -> None: ...
    def generate_from_2D_fractal_brownian_motion(self, in_value_one: Numerical, in_value_two: Numerical) -> None: ...
    def generate_from_3D_fractal_brownian_motion(self, in_value_one: Numerical, in_value_two: Numerical, in_value_three: Numerical) -> None: ...

    def set_percentage(self, value: Numerical) -> None: ...
    def set_decimal(self, value: Numerical) -> None: ...

    def get_percentage(self) -> Numerical: ...
    def get_decimal(self) -> Numerical: ...


class LinkedProportion(Proportion):
    def __init__(self, linked_class: Any, attr_name: str) -> None: ...


import threading
from typing import Any, Callable

class AdvancedThread(threading.Thread):
    def __init__(
        self,
        group: None = ...,
        target: Callable[..., Any] | None = ...,
        name: str | None = ...,
        args: tuple[Any, ...] = ...,
        kwargs: dict[str, Any] | None = ...,
        *,
        daemon: bool | None = ...,
    ) -> None: ...

    def kill(self) -> None: ...


from typing import Union

Numerical = Union[float, int]

class ButtonPressedEvent:
    def get_pressed(self) -> bool: ...
    def get_pressed_toggle(self) -> bool: ...
    def get_double_pressed(self) -> bool: ...
    def get_long_pressed(self) -> bool: ...

    def poll_long_pressed(self) -> bool: ...

    def get_repeat_press_duration(self) -> Numerical: ...
    def get_long_press_duration(self) -> Numerical: ...
    def get_double_press_duration(self) -> Numerical: ...

    def set_repeat_press_duration(self, duration: Numerical) -> None: ...
    def set_double_press_duration(self, duration: Numerical) -> None: ...
    def set_long_press_duration(self, duration: Numerical) -> None: ...

class Space:
    def __init__(self) -> None: ...

class Apostrophe(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Comma(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Minus(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Period(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Slash(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Zero(ButtonPressedEvent):
    def __init__(self) -> None: ...

class One(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Two(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Three(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Four(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Five(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Six(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Seven(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Eight(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Nine(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Semicolon(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Equal(ButtonPressedEvent):
    def __init__(self) -> None: ...

class A(ButtonPressedEvent):
    def __init__(self) -> None: ...

class B(ButtonPressedEvent):
    def __init__(self) -> None: ...

class C(ButtonPressedEvent):
    def __init__(self) -> None: ...

class D(ButtonPressedEvent):
    def __init__(self) -> None: ...

class E(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F(ButtonPressedEvent):
    def __init__(self) -> None: ...

class G(ButtonPressedEvent):
    def __init__(self) -> None: ...

class H(ButtonPressedEvent):
    def __init__(self) -> None: ...

class I(ButtonPressedEvent):
    def __init__(self) -> None: ...

class J(ButtonPressedEvent):
    def __init__(self) -> None: ...

class K(ButtonPressedEvent):
    def __init__(self) -> None: ...

class L(ButtonPressedEvent):
    def __init__(self) -> None: ...

class M(ButtonPressedEvent):
    def __init__(self) -> None: ...

class N(ButtonPressedEvent):
    def __init__(self) -> None: ...

class O(ButtonPressedEvent):
    def __init__(self) -> None: ...

class P(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Q(ButtonPressedEvent):
    def __init__(self) -> None: ...

class R(ButtonPressedEvent):
    def __init__(self) -> None: ...

class S(ButtonPressedEvent):
    def __init__(self) -> None: ...

class T(ButtonPressedEvent):
    def __init__(self) -> None: ...

class U(ButtonPressedEvent):
    def __init__(self) -> None: ...

class V(ButtonPressedEvent):
    def __init__(self) -> None: ...

class W(ButtonPressedEvent):
    def __init__(self) -> None: ...

class X(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Y(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Z(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Left_Bracket(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Backslash(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Right_Bracket(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Grave_Accent(ButtonPressedEvent):
    def __init__(self) -> None: ...

class World_1(ButtonPressedEvent):
    def __init__(self) -> None: ...

class World_2(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Escape(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Enter(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Tab(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Backspace(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Insert(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Delete(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Right(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Left(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Down(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Up(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Page_Up(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Page_Down(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Home(ButtonPressedEvent):
    def __init__(self) -> None: ...

class End(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Caps_Lock(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Scroll_Lock(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Num_Lock(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Print_Screen(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Pause(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F1(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F2(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F3(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F4(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F5(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F6(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F7(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F8(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F9(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F10(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F11(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F12(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F13(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F14(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F15(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F16(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F17(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F18(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F19(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F20(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F21(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F22(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F23(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F24(ButtonPressedEvent):
    def __init__(self) -> None: ...

class F25(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Left_Shift(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Left_Control(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Left_Alt(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Left_Super(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Right_Shift(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Right_Control(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Right_Alt(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Right_Super(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Shift(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Control(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Alt(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Super(ButtonPressedEvent):
    def __init__(self) -> None: ...

class Menu(ButtonPressedEvent):
    def __init__(self) -> None: ...


class Passport:
    def set_product_name(self, product_name: str) -> None: ...
    def set_product_sub_name(self, product_sub_name: str) -> None: ...
    def set_company_name(self, company_name: str) -> None: ...
    def set_product_version(self, product_version: str) -> None: ...
    def set_product_path(self, product_path: str) -> None: ...
    def set_logging_path(self, logging_path: str) -> None: ...
    def set_profiling_path(self, profiling_path: str) -> None: ...

    def register(self) -> None: ...

    def is_registered(self) -> bool: ...

    def get_product_name(self) -> str: ...
    def get_product_sub_name(self) -> str: ...
    def get_company_name(self) -> str: ...
    def get_product_version(self) -> str: ...
    def get_product_path(self) -> str: ...
    def get_logging_path(self) -> str: ...
    def get_profiling_path(self) -> str: ...


from typing import Literal

class Tkinter:
    @staticmethod
    def style(widget: Literal[
            "TButton", "TCheckbutton", "TCombobox", "TEntry", "TFrame",
            "TLabel", "TLabelFrame", "TMenubutton", "TNotebook",
            "TPanedwindow", "Horizontal.TProgressbar",
            "Vertical.TProgressbar", "TRadiobutton", "Horizontal.TScale",
            "Vertical.TScale", "Horizontal.TScrollbar",
            "Vertical.TScrollbar", "TSeparator", "TSizegrip",
            "Treeview"]) -> None: ...


from typing import Iterable, Union

import numpy as np
import numpy.typing as npt

Numerical = Union[float, int]

Numerical1D = Union[
    npt.NDArray[np.float32], # preferred
    npt.NDArray[np.float16],
    npt.NDArray[np.float64],
    npt.NDArray[np.int32], # preferred
    npt.NDArray[np.int8],
    npt.NDArray[np.int16],
    npt.NDArray[np.int64],
    Iterable[float],  # preferred
    Iterable[int],  # preferred
]

class ButtonPressedEvent:
    def get_pressed(self) -> bool: ...
    def get_pressed_toggle(self) -> bool: ...
    def get_double_pressed(self) -> bool: ...
    def get_long_pressed(self) -> bool: ...

    def poll_long_pressed(self) -> bool: ...

    def get_repeat_press_duration(self) -> Numerical: ...
    def get_long_press_duration(self) -> Numerical: ...
    def get_double_press_duration(self) -> Numerical: ...

    def set_repeat_press_duration(self, duration: Numerical) -> None: ...
    def set_double_press_duration(self, duration: Numerical) -> None: ...
    def set_long_press_duration(self, duration: Numerical) -> None: ...

class Controller:
    a_button: ButtonPressedEvent
    b_button: ButtonPressedEvent
    x_button: ButtonPressedEvent
    y_button: ButtonPressedEvent
    left_bumper_button: ButtonPressedEvent
    right_bumper_button: ButtonPressedEvent
    back_button: ButtonPressedEvent
    start_button: ButtonPressedEvent
    guide_button: ButtonPressedEvent
    left_thumb_button: ButtonPressedEvent
    right_thumb_button: ButtonPressedEvent
    dpad_up_button: ButtonPressedEvent
    dpad_down_button: ButtonPressedEvent
    dpad_left_button: ButtonPressedEvent
    dpad_right_button: ButtonPressedEvent

    def __init__(self, controller_id: int) -> None: ...

    def get_connected(self) -> bool: ...
    def get_active(self) -> bool: ...

    def get_raw_name(self) -> bool: ...
    def get_gamepad_name(self) -> bool: ...
    def get_GUID(self) -> bool: ...

    def get_axis_decimal(self, axis_id: int) -> Numerical: ...
    def get_axis_percentage(self, axis_id: int) -> Numerical: ...

    def get_raw_button_pressed(self, button_id: int) -> bool: ...

    def get_raw_hat_state(self, hat_id: int) -> str: ...

    def get_right_stick_X_axis_percentage(self) -> Numerical: ...
    def get_right_stick_Y_axis_percentage(self) -> Numerical: ...

    def get_right_stick_X_axis_decimal(self) -> Numerical: ...
    def get_right_stick_Y_axis_decimal(self) -> Numerical: ...

    def get_left_stick_X_axis_percentage(self) -> Numerical: ...
    def get_Left_stick_Y_axis_percentage(self) -> Numerical: ...

    def get_left_stick_X_axis_decimal(self) -> Numerical: ...
    def get_Left_stick_Y_axis_decimal(self) -> Numerical: ...

    def get_right_trigger_axis_percentage(self) -> Numerical: ...
    def get_left_trigger_axis_percentage(self) -> Numerical: ...

    def get_right_trigger_axis_decimal(self) -> Numerical: ...
    def get_left_trigger_axis_decimal(self) -> Numerical: ...

    def get_left_stick_position_percentage(self, dead_zone: float=0, using_numpy_arrays: bool=True) -> Numerical1D: ...
    def get_right_stick_position_percentage(self, dead_zone: float=0, using_numpy_arrays: bool=True) -> Numerical1D: ...

    def get_left_stick_position_decimal(self, dead_zone: float=0, using_numpy_arrays: bool=True) -> Numerical1D: ...
    def get_right_stick_position_decimal(self, dead_zone: float=0, using_numpy_arrays: bool=True) -> Numerical1D: ...


class TimeFormatter:
    def set_from_year(self, year: float) -> None: ...
    def set_from_month(self, month: float) -> None: ...
    def set_from_day(self, day: float) -> None: ...
    def set_from_hour(self, hour: float) -> None: ...
    def set_from_minute(self, minute: float) -> None: ...
    def set_from_second(self, second: float) -> None: ...
    def set_from_microsecond(self, microsecond: float) -> None: ...

    def get_year(self) -> float: ...
    def get_month(self) -> float: ...
    def get_day(self) -> float: ...
    def get_hour(self) -> float: ...
    def get_minute(self) -> float: ...
    def get_second(self) -> float: ...
    def get_microsecond(self) -> float: ...

    def get_in_date_format(self) -> str: ...
    def get_in_sentence_format(self) -> str: ...


from typing import Iterable, Union

import numpy as np
import numpy.typing as npt

Numerical1D = Union[
    npt.NDArray[np.float32], # preferred
    npt.NDArray[np.float16],
    npt.NDArray[np.float64],
    npt.NDArray[np.int32], # preferred
    npt.NDArray[np.int8],
    npt.NDArray[np.int16],
    npt.NDArray[np.int64],
    Iterable[float],  # preferred
    Iterable[int],  # preferred
]

Float1D = Union[npt.NDArray[np.float32], Iterable[float]]

Numerical = Union[float, int]

class AdvancedMathematics:
    @staticmethod
    def lerp(x1: float, x2: float, duration: float, current_duration: float) -> float: ...

    @staticmethod
    def individual_pythagorean_difference(x1: Numerical, y1: Numerical, x2: Numerical, y2: Numerical) -> float: ...

    @staticmethod
    def point_pythagorean_difference(point1: Numerical1D, point2: Numerical1D) -> float: ...

    @staticmethod
    def individual_pythagorean_distance(x: Numerical, y: Numerical) -> float: ...

    @staticmethod
    def point_pythagorean_distance(point: Numerical1D) -> float: ...

    @staticmethod
    def smooth_step(value: Numerical) -> float: ...

    @staticmethod
    def ranger(value: Numerical, old_range: Numerical1D, new_range: Numerical1D) -> float: ...

    @staticmethod
    def array_ranger(values: Numerical1D, old_range: Numerical1D, new_range: Numerical1D) -> Numerical1D: ...

    @staticmethod
    def array_normalize(values: Numerical) -> Numerical1D: ...

    @staticmethod
    def cross(first_values: Numerical1D, second_values: Numerical1D) -> Numerical1D: ...

    @staticmethod
    def subtract(first_values: Numerical1D, second_values: Numerical1D) -> Numerical1D: ...

    @staticmethod
    def dot(first_values: Numerical1D, second_values: Numerical1D) -> float: ...

    @staticmethod
    def look_at(eye: Numerical1D, target: Numerical1D, up: Numerical1D) -> Numerical1D: ...

    @staticmethod
    def compute_position(position: Numerical1D, target: Numerical1D, up: Numerical1D) -> Numerical1D: ...

    @staticmethod
    def perspective_fov(FOV: Numerical, aspect_ratio: Numerical, near_plane: Numerical, far_plane: Numerical) -> Numerical1D: ...


from typing import Any, Iterable

Any1D = Iterable[Any]

class Logger:
    def is_log_to_file(self) -> bool: ...
    def is_log_to_console(self) -> bool: ...

    def set_log_to_file(self, log_to_file: bool) -> None: ...
    def set_log_to_console(self, log_to_console: bool) -> None: ...
    def set_max_log_files_to_keep(self, value: int) -> None: ...
    def set_log_debug(self, log_debug: bool) -> None: ...
    def set_log_warn(self, log_warn: bool) -> None: ...
    def set_log_error(self, log_error: bool) -> None: ...

    def get_log_debug(self) -> bool: ...
    def get_log_warn(self) -> bool: ...
    def get_log_error(self) -> bool: ...
    def get_max_log_files_to_keep(self) -> int: ...

    def log_debug_with_id(
        self, log_id: str, content: str, product_name: str="",
        repeat_for_effect: bool=False) -> None: ...
    def log_warn_with_id(
        self, log_id: str, content: str, product_name: str="",
        repeat_for_effect: bool=False) -> None: ...
    def log_error_with_id(
        self, log_id: str, content: str, product_name: str="",
        repeat_for_effect: bool=False) -> None: ...

    def log_debug(
        self, content: str, values: Any1D=[], product_name: str="",
        repeat_for_effect: bool=False) -> None: ...
    def log_warn(
        self, content: str, values: Any1D=[], product_name: str="",
        repeat_for_effect: bool=False) -> None: ...
    def log_error(
        self, content: str, values: Any1D=[], product_name: str="",
        repeat_for_effect: bool=False) -> None: ...


from typing import Union, Iterable

import numpy as np
import numpy.typing as npt

Numerical = Union[float, int]
NoneInt = Union[None, int]

Numerical1D = Union[
    npt.NDArray[np.float32], # preferred
    npt.NDArray[np.float16],
    npt.NDArray[np.float64],
    npt.NDArray[np.int32], # preferred
    npt.NDArray[np.int8],
    npt.NDArray[np.int16],
    npt.NDArray[np.int64],
    Iterable[float],  # preferred
    Iterable[int],  # preferred
]

Numerical2D = Union[
    npt.NDArray[np.float32], # preferred
    npt.NDArray[np.float16],
    npt.NDArray[np.float64],
    npt.NDArray[np.int32], # preferred
    npt.NDArray[np.int8],
    npt.NDArray[np.int16],
    npt.NDArray[np.int64],
    Iterable[Iterable[float]],  # preferred
    Iterable[Iterable[int]],  # preferred
]

Numerical3D = Union[
    npt.NDArray[np.float32], # preferred
    npt.NDArray[np.float16],
    npt.NDArray[np.float64],
    npt.NDArray[np.int32], # preferred
    npt.NDArray[np.int8],
    npt.NDArray[np.int16],
    npt.NDArray[np.int64],
    Iterable[Iterable[Iterable[float]]],  # preferred
    Iterable[Iterable[Iterable[int]]],  # preferred
]

class FractalBrownianMotion:
    def __init__(self, seed: NoneInt=None, octaves: int=2, lacunarity: float=0.75, gain: float=1.0) -> None: ...

    def get_seed(self) -> int: ...
    def get_octaves(self) -> int: ...
    def get_lacunarity(self) -> float: ...
    def get_gain(self) -> float: ...

    def noise_1d(self, x: Numerical) -> float: ...
    def noise_2d(self, x: Numerical, y: Numerical) -> float: ...
    def noise_3d(self, x: Numerical, y: Numerical, z: Numerical) -> float: ...

    def array_noise_1d(self, values: Numerical1D) -> Numerical1D: ...
    def array_noise_2d(self, values: Numerical2D) -> Numerical2D: ...
    def array_noise_3d(self, values: Numerical3D) -> Numerical3D: ...

    def range_noise_1d(self, x_range: Numerical1D, length: Numerical) -> Numerical1D: ...
    def range_noise_2d(self, x_range: Numerical1D, y_range: Numerical1D, length: Numerical) -> Numerical1D: ...
    def range_noise_3d(self, x_range: Numerical1D, y_range: Numerical1D, z_range: Numerical1D, length: Numerical) -> Numerical1D: ...


class NoInputDevicesFoundError(Exception):
    """No input devices were found!"""
    pass

class UnableToReadAudioSampleError(Exception):
    """Unable to read audio sample!"""
    pass

class PassportNotInitializedError(Exception):
    """Passport not initialized!"""
    pass

class PassportNotRegisteredError(Exception):
    """Passport has not been registered!"""
    pass

class DirectoryNotFoundError(Exception):
    """Directory not found!"""
    pass

class DirectoryAlreadyExistsError(Exception):
    """Directory already exists!"""
    pass


#type: ignore

from setuptools import setup, Extension
from Cython.Build import cythonize
import sys, os
import platform
import multiprocessing

import numpy

cwd = os.path.dirname(__file__)

sys.path.insert(0, os.path.join(cwd, "pmma", "build"))

def add_source(name: str):
    return [
        os.path.join("pmma", "core", "pyx_src", f"{name}.pyx"),
        os.path.join("pmma", "core", "cpp_src", f"{name}.cpp")
        ]

if sys.platform.startswith("win"):
    compile_args = ["/O2", "/fp:fast", "/GL", "/GF", "/GS-", "/std:c++17", "/wd4551", "/wd4251"] # disable warning 4551 & 4251 which is an issue for Cython
    link_args = ["/LTCG"]

elif sys.platform.startswith("linux"):
    compile_args = [
        "-O3", "-ffast-math", "-funroll-loops", "-fstrict-aliasing", "-fomit-frame-pointer", "-std=c++17"
    ]
    link_args = []

elif sys.platform == "darwin":
    compile_args = [
        "-O3", "-ffast-math", "-funroll-loops", "-fstrict-aliasing", "-fomit-frame-pointer", "-std=c++17"
    ]
    link_args = [
        "-framework", "Cocoa",
        "-framework", "OpenGL",
        "-framework", "IOKit",
        "-framework", "CoreVideo",
    ]

    arch = platform.machine()
    if arch == "arm64":
        compile_args.append("-arch")
        compile_args.append("arm64")
        link_args.append("-arch")
        link_args.append("arm64")
else:
    raise NotImplementedError("Unsupported platform")

shared_name = 'PMMA_Core'

def make_ext(component, extra_cpp=None, add_numpy=False, raw_depends=[]):
    sources = [os.path.join("pmma", "core", "pyx_src", component)]
    if extra_cpp is not None:
        sources.extend(extra_cpp)

    depends = []
    for item in raw_depends:
        depends.append(os.path.join("pmma", "core", "pyx_src", item))

    lib_dirs = [os.path.join(cwd, "pmma", "lib")]

    libs = [shared_name]

    includes = [os.path.join(cwd, "pmma", "core", "hpp_src"), os.path.join(cwd, "pmma", "extern", "include")]

    if add_numpy:
        includes += [numpy.get_include()]

    macros = []
    if add_numpy:
        macros.append(('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION'))

    name = component.split(os.sep)[-1].replace(".pyx", "")

    return Extension(
        name=name,
        sources=sources,
        language="c++",
        include_dirs=includes,
        library_dirs=lib_dirs,
        libraries=libs,
        extra_compile_args=compile_args,
        extra_link_args=link_args,
        define_macros=macros,
        depends=depends
    )

ext_modules = [
    make_ext("AdvancedMathematics.pyx", add_numpy=True),
    make_ext("Display.pyx", add_numpy=True, raw_depends=["NumberFormats.pyx"]),
    make_ext(os.path.join("Events", "KeyEvents.pyx")),
    make_ext(os.path.join("Events", "KeyPadEvents.pyx")),
    make_ext(os.path.join("Events", "WindowEvents.pyx")),
    make_ext(os.path.join("Events", "ControllerEvents.pyx"), add_numpy=True),
    make_ext(os.path.join("Events", "MouseEvents.pyx"), add_numpy=True),
    make_ext("FractalBrownianMotion.pyx", add_numpy=True),
    make_ext("NumberFormats.pyx", add_numpy=True),
    make_ext("PerlinNoise.pyx", add_numpy=True),
    make_ext("PMMA_Core.pyx", raw_depends=["General.pyx"]),
    make_ext("TextRenderer.pyx", add_numpy=True),
    make_ext("General.pyx", add_numpy=True),
    make_ext("Shapes2D.pyx", add_numpy=True, raw_depends=["NumberFormats.pyx"]),
    make_ext("Passport.pyx", raw_depends=["General.pyx", "Logger.pyx"]),
    make_ext("Logger.pyx"),
    make_ext("Animation.pyx", add_numpy=True, raw_depends=["NumberFormats.pyx"])
]

# Read the long description from README.md
with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# Read the requirements from requirements.txt
with open("requirements.txt", "r", encoding="utf-8") as req_file:
    requirements = req_file.read().splitlines()

packages = ['pmma']

if __name__ == '__main__':
    multiprocessing.freeze_support()

    use_parallel = True
    if '--no-parallel' in sys.argv:
        use_parallel = False
        sys.argv.remove('--no-parallel')

    annotate_build = "--annotate_build" in sys.argv
    if annotate_build:
        sys.argv.remove("--annotate_build")

    if use_parallel:
        cython_command = cythonize(ext_modules, compiler_directives={"language_level": "3"}, annotate=annotate_build, nthreads=multiprocessing.cpu_count())
    else:
        cython_command = cythonize(ext_modules, compiler_directives={"language_level": "3"}, annotate=annotate_build)

    setup(
        name="pmma",
        version="5.0.15",
        author="PycraftDev",
        author_email="thomasjebbo@gmail.com",
        description="Python Multi-Media API (PMMA) is a multi-purpose API designed to make working on multi-media projects easier and faster!",
        long_description=long_description,
        long_description_content_type="text/markdown",
        url="https://github.com/PycraftDeveloper/PMMA",
        project_urls={
            "Bug Tracker": "https://github.com/PycraftDeveloper/PMMA/issues",
        },
        classifiers=[
            "Programming Language :: Python :: 3",
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
        ],
        packages=packages,  # Include the pmma package and all its sub-packages
        python_requires=">=3.8",
        install_requires=requirements,
        include_package_data=True,
        ext_modules=cython_command, # nthreads=multiprocessing.cpu_count()
    )


import threading

class Registry:
    render_thread = None
    checking_for_updates = True
    update_checking_thread = None
    passport_instance = None
    profiler_instance = None


def require_render_thread(func):
    def wrapper(*args, **kwargs):
        if threading.current_thread() != Registry.render_thread:
            raise RuntimeError(f"{func.__name__} must be called from the render thread.")
        return func(*args, **kwargs)
    return wrapper


import os
import platform
import ctypes
import sys
import threading

system = platform.system()

pmma_dir = os.path.dirname(os.path.abspath(__file__))

sys.path.append(os.path.join(pmma_dir, "build"))

pmma_lib_dir = os.path.join(pmma_dir, "lib")

if system == "Windows":
    if "PMMA_DEBUG" in os.environ and os.environ["PMMA_DEBUG"] == "1":
        try:
            extern_binaries = os.path.join(pmma_dir, "extern - DEBUG", "bin")
            ctypes.CDLL(os.path.join(extern_binaries, "zlibd.dll"))
            ctypes.CDLL(os.path.join(extern_binaries, "libpng16d.dll"))
            print("You are using a DEBUGGING version of PMMA, this is not \
meant for prime time and is used only for debugging or development testing.")
        except FileNotFoundError:
            print("Could not find DEBUG files, please run: 'build_tools/main - debug.py'. Using release files instead")
            extern_binaries = os.path.join(pmma_dir, "extern", "bin")
            ctypes.CDLL(os.path.join(extern_binaries, "zlib.dll"))
            ctypes.CDLL(os.path.join(extern_binaries, "libpng16.dll"))
    else:
        extern_binaries = os.path.join(pmma_dir, "extern", "bin")
        ctypes.CDLL(os.path.join(extern_binaries, "zlib.dll"))
        ctypes.CDLL(os.path.join(extern_binaries, "libpng16.dll"))

    ctypes.CDLL(os.path.join(pmma_lib_dir, "PMMA_Core.dll"))

elif system == "Linux":
    if "PMMA_DEBUG" in os.environ and os.environ["PMMA_DEBUG"] == "1":
        print("We don't currently support Linux debugging. You might be \
the person to help change this! Defaulting to release files instead.")

    extern_libs = os.path.join(pmma_dir, "extern", "lib")

    ctypes.CDLL(os.path.join(extern_libs, "libz.so"))
    ctypes.CDLL(os.path.join(extern_libs, "libpng16.so"))
    ctypes.CDLL(os.path.join(pmma_lib_dir, "libPMMA_Core.so"))

elif system == "Darwin":
    if "PMMA_DEBUG" in os.environ and os.environ["PMMA_DEBUG"] == "1":
        print("We don't currently support Apple platform debugging. You \
might be the person to help change this! Defaulting to release files \
instead.")

    extern_libs = os.path.join(pmma_dir, "extern", "lib")

    ctypes.CDLL(os.path.join(extern_libs, "zlib.dylib"))
    ctypes.CDLL(os.path.join(extern_libs, "libpng16.dylib"))
    ctypes.CDLL(os.path.join(pmma_lib_dir, "libPMMA_Core.dylib"))

import pmma.build.PMMA_Core as _PMMA_Core
_PMMA_Core.initialize()

from pmma.build.General import General

General.set_pmma_location(pmma_dir)
General.set_path_separator()

from pmma.core.py_src.Utility import Registry as _Registry
_Registry.update_checking_thread = threading.Thread(target=General.internal_check_for_updates, daemon=True)
_Registry.update_checking_thread.start()

from pmma.build.AdvancedMathematics import AdvancedMathematics
from pmma.build.PerlinNoise import PerlinNoise
from pmma.build.FractalBrownianMotion import FractalBrownianMotion
from pmma.build.Display import Display
import pmma.build.NumberFormats as NumberFormats
import pmma.build.KeyEvents as KeyEvents
import pmma.build.KeyPadEvents as KeyPadEvents
import pmma.build.WindowEvents as WindowEvents
import pmma.build.ControllerEvents as ControllerEvents
import pmma.build.MouseEvents as MouseEvents
from pmma.build.TextRenderer import *
import pmma.build.Shapes2D as Shapes2D
from pmma.build.Passport import Passport
from pmma.build.Logger import Logger
import pmma.build.Animation as Animation

from pmma.core.py_src.Backpack import Backpack
from pmma.core.py_src.Audio import *
from pmma.core.py_src.Executor import Executor, AdvancedExecutor
from pmma.core.py_src.GPU import GPUs
from pmma.core.py_src.AdvancedThreading import AdvancedThread
import pmma.core.py_src.DataStructures as DataStructures
from pmma.core.py_src.Sampler import Sampler
from pmma.core.py_src.DateTimeFormatter import TimeFormatter
import pmma.core.py_src.FileManagement as FileManagement